import json

from .forms import MessageForm
from .forms import message_fields

from . import exceptions


__all__ = [
    'Base',
    'Request',
    'Response',
    'Error',
]

class Base(dict):

    name = None

    class Form(MessageForm):
        fields = [
            message_fields.name__required,
            message_fields.body,
        ]

    def __init__(self, **kwargs):
        kwargs = self.Form().to_python_or_exc(kwargs)
        if self.name and kwargs['name']!=self.name:
            raise MessageError(errors={'name':'Name error'})
        super().__init__(**kwargs)


class Request(Base):

    name = 'request'

    class Form(MessageForm):
        fields = [
            message_fields.name__required,
            message_fields.request_id__required,
            message_fields.handler__required,
            message_fields.body,
        ]


class Response(Base):

    name = 'response'

    class Form(MessageForm):
        fields = [
            message_fields.name__required,
            message_fields.request_id__required,
            message_fields.handler__required,
            message_fields.body,
        ]


class Error(Base):

    name = 'error'

    class Form(MessageForm):
        fields = [
            message_fields.name__required,
            message_fields.request_id,
            message_fields.handler,
            message_fields.body__error_required,
        ]




import os

import pkg_resources
from six.moves.urllib.parse import urlparse


class UrlBase(object):
    def __init__(self, url):
        super(UrlBase, self).__init__()
        self._url = url

    def isdir(self):
        raise NotImplementedError

    def isfile(self):
        raise NotImplementedError

    def exists(self):
        raise NotImplementedError

    def open_file(self, mode='r'):
        raise NotImplementedError

    def read(self):
        raise NotImplementedError

    def listdir(self):
        raise NotImplementedError

    def dirname(self):
        raise NotImplementedError

    def makedirs(self):
        raise NotImplementedError

    def isreadonly(self):
        raise NotImplementedError

    def join(self, path):
        raise NotImplementedError

    def __str__(self):
        return self._url.geturl()


class FSUrl(UrlBase):
    def __init__(self, root, url):
        super(FSUrl, self).__init__(url)
        self.root = root
        if self._url.netloc:
            raise ValueError('Netloc not allowed', str(self))
        if self._url.params:
            raise ValueError('Params not allowed', str(self))
        if self._url.query:
            raise ValueError('Query not allowed', str(self))
        if self._url.fragment:
            raise ValueError('Fragment not allowed', str(self))
        self.path = os.path.join(root, url.path)

    def isdir(self):
        return os.path.isdir(self.path)

    def isfile(self):
        return os.path.isfile(self.path)

    def exists(self):
        return os.path.exists(self.path)

    def open_file(self, mode='r'):
        return open(self.path, mode)

    def read(self):
        with self.open() as f:
            return f.read()

    def listdir(self):
        return os.path.listdir(self.path)

    def dirname(self):
        return os.path.dirname(self.path)

    def makedirs(self):
        if not self.exists():
            os.makedirs(self.path)

    def isreadonly(self):
        # XXX
        return False

    def join(self, path):
        return os.path.join(self.path, path)

    def __str__(self):
        return self.path


class PKGResourcesUrl(UrlBase):
    def __init__(self, url):
        super(PKGResourcesUrl, self).__init__(url)
        if self._url.params:
            raise ValueError('Params not allowed', str(self))
        if self._url.query:
            raise ValueError('Query not allowed', str(self))
        if self._url.fragment:
            raise ValueError('Fragment not allowed', str(self))
        self.package = url.netloc
        self.path = url.path

    def isdir(self):
        return pkg_resources.resource_isdir(self.package, self.path)

    def isfile(self):
        return self.exists() and not self.isdir()

    def exists(self):
        return pkg_resources.resource_exists(self.package, self.path)

    def open_file(self, mode='r'):
        if mode not in ('r', 'rb', 'U'):
            raise ValueError('File is readonly', str(self))
        return pkg_resources.resource_stream(self.package, self.path)

    def read(self):
        with self.open() as f:
            return f.read()

    def listdir(self):
        return pkg_resources.resource_listdir(self.package, self.path)

    def isreadonly(self):
        return True


SCHEMES = {
    '': lambda cfg, url: FSUrl(cfg.ROOT_DIR, url),
    'pkg': lambda cfg, url: PKGResourcesUrl(url),
}


class PathBase(object):
    def __init__(self, cfg, str_path, schemes=None):
        super(PathBase, self).__init__()
        self.cfg = cfg
        url = urlparse(str_path)
        if schemes is None:
            schemes = list(SCHEMES)
        if url.scheme not in schemes or url.scheme not in SCHEMES:
            raise ValueError(
                'Schema {} not allowed'.format(url.scheme),
                str(self),
            )
        self.url = SCHEMES[url.scheme](cfg, url)
        self.scheme = url.scheme

    def __str__(self):
        return str(self.url)


class DirPath(PathBase):
    def __init__(self, cfg, str_path, schemes=['']):
        super(DirPath, self).__init__(cfg, str_path, schemes)

    def check(self):
        if not self.url.exists():
            raise ValueError('Dir not exists', str(self))
        if not self.url.isdir():
            raise ValueError('Is not dir', str(self))

    def exists(self):
        return self.url.exists()

    def listdir(self):
        return self.url.listdir()

    def join(self, path):
        return self.url.join(path)

    def makedirs(self):
        self.url.makedirs()


class FilePath(PathBase):
    def __init__(self, cfg, str_path, schemes=['']):
        super(FilePath, self).__init__(cfg, str_path, schemes)

    def check(self):
        if not self.url.exists():
            raise ValueError('File not exists', str(self))
        if not self.url.isfile():
            raise ValueError('Is not file', str(self))

    def exists(self):
        return self.url.exists()

    def open(self, mode='r'):
        return self.url.open_file(mode)

    def read(self):
        return self.url.read()

    def dirname(self):
        return DirPath(self.cfg, self.url.dirname())

    def makedirs(self):
        self.dirname().makedirs()

    def isreadonly(self):
        return self.url.isreadonly()



import math

from ikcms.utils import cached_property


class PageNotFound(Exception):
    pass


class Paginator(object):

    class Page(object):

        def __init__(self, paginator, number, items):
            self.paginator = paginator
            self.count = self.paginator.count
            self.number = number
            self.items = items
            self.first = bool(number == 1)
            self.last = bool(number == paginator.pages_count)
            self.has_prev = not self.first
            self.has_next = not self.last
            self.next_page_number = self.has_next and (self.number + 1) or None
            self.prev_page_number = self.has_prev and (self.number - 1) or None

        def __iter__(self):
            return self.items.__iter__()

        @cached_property
        def next_page(self):
            if self.has_next:
                return self.paginator.page(self.number + 1)

        @cached_property
        def prev_page(self):
            if self.has_prev:
                return self.paginator.page(self.number - 1)



    def __init__(self, query, limit):
        self.query = query
        self.limit = limit
        self.count = query.count()
        self.pages_count = int(math.ceil(float(self.count or 1) / self.limit))

    def page(self, page):
        try:
            page = int(page)
            if (page < 1) or (page > self.pages_count):
                raise ValueError
        except (TypeError, ValueError):
            raise PageNotFound(page)
        return self.Page(self, page, self._get_items(page))

    def _get_items(self, page):
        return self.query[self.limit * (page - 1):self.limit * page]




from sqlalchemy import exc
from sqlalchemy.dialects.mysql.pymysql import dialect
from sqlalchemy.sql import compiler
from sqlalchemy.sql import expression
from sqlalchemy.sql import elements
from sqlalchemy.sql import selectable

from .expression import escape


class SQLCompiler(compiler.SQLCompiler):

    def visit_column(self, column, result_map=None, **kwargs):
        name = column.name

        if name is None:
            raise exc.CompileError(
                'Cannot compile Column object until it\'s "name" is assigned.'
            )

        is_literal = column.is_literal
        if not is_literal and isinstance(name, elements._truncated_label):
            name = self._truncated_identifier("colident", name)

        #if result_map is not None:
        #    result_map[name.lower()] = (name, (column, ), column.type)

        if is_literal:
            name = self.escape_literal_column(name)
        else:
            name = self.preparer.quote(name, column.name)

        return name

    def visit_select(
            self,
            select,
            asfrom=False,
            parens=True,
            iswrapper=False,
            fromhints=None,
            compound_index=0,
            force_result_map=False,
            nested_join_translation=False,
            **kwargs
    ):
        needs_nested_translation = \
            select.use_labels and \
            not nested_join_translation and \
            not self.stack and \
            not self.dialect.supports_right_nested_joins

        if needs_nested_translation:
            transformed_select = self._transform_select_for_nested_joins(
                select)
            text = self.visit_select(
                transformed_select, asfrom=asfrom, parens=parens,
                iswrapper=iswrapper, fromhints=fromhints,
                compound_index=compound_index,
                force_result_map=force_result_map,
                nested_join_translation=True, **kwargs
            )

        toplevel = not self.stack
        entry = self._default_stack_entry if toplevel else self.stack[-1]

        populate_result_map = force_result_map or (
            compound_index == 0 and (
                toplevel or
                entry['iswrapper']
            )
        )

        if needs_nested_translation:
            if populate_result_map:
                self._transform_result_map_for_nested_joins(
                    select, transformed_select)
            return text

        correlate_froms = entry['correlate_froms']
        asfrom_froms = entry['asfrom_froms']

        if asfrom:
            froms = select._get_display_froms(
                explicit_correlate_froms=correlate_froms.difference(
                    asfrom_froms),
                implicit_correlate_froms=())
        else:
            froms = select._get_display_froms(
                explicit_correlate_froms=correlate_froms,
                implicit_correlate_froms=asfrom_froms)

        new_correlate_froms = set(selectable._from_objects(*froms))
        all_correlate_froms = new_correlate_froms.union(correlate_froms)

        new_entry = {
            'asfrom_froms': new_correlate_froms,
            'iswrapper': iswrapper,
            'correlate_froms': all_correlate_froms,
            'selectable': select,
        }
        self.stack.append(new_entry)

        column_clause_args = kwargs.copy()
        column_clause_args.update({
            'within_label_clause': False,
            'within_columns_clause': False
        })

        text = "SELECT "

        # Add DISTINCT keyword if necessary
        text += self.get_select_precolumns(select)

        # List of columns to print in the SELECT column list.
        inner_columns = [
            c for c in [
                self._label_select_column(select,
                                          column,
                                          populate_result_map, asfrom,
                                          column_clause_args,
                                          name=name)
                for name, column in select._columns_plus_names
            ]
            if c is not None
        ]

        text += ', '.join(inner_columns)

        if froms:
            text += " \nFROM "
            text += ', '.join(
                [
                    f._compiler_dispatch(self, asfrom=True, **kwargs)
                    for f in froms
                ]
            )
        else:
            text += self.default_from()

        if select._whereclause is not None:
            t = select._whereclause._compiler_dispatch(self, **kwargs)
            if t:
                text += " \nWHERE " + t

        if select._group_by_clause.clauses:
            group_by = select._group_by_clause._compiler_dispatch(
                self, **kwargs)
            if group_by:
                text += " GROUP BY " + group_by

        if select._having is not None:
            t = select._having._compiler_dispatch(self, **kwargs)
            if t:
                text += " \nHAVING " + t

        if select._order_by_clause.clauses:
            text += self.order_by_clause(select, **kwargs)

        if select._limit is not None:
            text += self.limit_clause(select)

        if getattr(select, '_options', None) is not None:
            text += self.options_clause(select)

        self.stack.pop(-1)

        if asfrom and parens:
            return "(" + text + ")"
        else:
            return text

    def options_clause(self, select):
        def proccess(options, parenthesis=False):
            result = ', '.join([
                '{}={}'.format(
                    k,
                    v if not isinstance(v, dict) else proccess(v.items(), True)
                )
                for k, v in options
            ])
            if parenthesis:
                return '({})'.format(result)
            return result
        return '\nOPTION {}'.format(proccess(select._options))

    def limit_clause(self, select):
        limit, offset = select._limit, select._offset
        if limit:
            if offset:
                return ' \nLIMIT %s, %s' % (
                    self.process(expression.literal(offset)),
                    self.process(expression.literal(limit)))
            return ' \nLIMIT %s' % self.process(expression.literal(limit))
        return ''

    def visit_multi(self, element, **kwargs):
        return '(%s)' % self.visit_clauselist(element, **kwargs)

    def visit_match(self, element, **kwargs):
        text = escape(' '.join(element.values))
        text += ' '.join(u'@{} {}'.format(k, escape(v)) for k, v in element.fields.items())
        return 'match(%s)' % self.process(expression.literal(text))


class Dialect(dialect):
    name = 'sphinxql'
    default_paramstyle = 'format'
    positional = True
    ddl_compiler = None
    statement_compiler = SQLCompiler
    supports_unicode_statements = True
    supports_multivalues_insert = True
    supports_right_nested_joins = False
    supports_alter = False
    supports_views = False
    description_encoding = None

    def _get_server_version_info(self, connection):
        return (4, 1)

    def _check_unicode_returns(self, connection, additional_tests=None):
        return True

    def _detect_ansiquotes(self, connection):
        self._server_ansiquotes = False
        self._backslash_escapes = False

    def _get_default_schema_name(self, connection):
        raise NotImplementedError

    def get_isolation_level(self, connection):
        raise NotImplementedError



from ikcms.ws_apps.base.exceptions import BaseError
from ikcms.ws_apps.base.exceptions import ClientError
from ikcms.ws_apps.base.exceptions import MessageError
from ikcms.forms.exceptions import ValidationError
from ikcms.forms.exceptions import RawValueError
from ikcms.forms.exceptions import RawValueTypeError
from ikcms.ws_components.auth.exceptions import AccessDeniedError


__all__ = (
    'ClientError',
    'MessageError',
    'ValidationError',
    'RawValueError',
    'RawValueTypeError',
    'AccessDeniedError',
    'StreamBaseError',
    'StreamNotFound',
    'StreamActionNotFound',
    'StreamFieldNotFound',
    'StreamItemNotFoundError',
    'StreamItemAlreadyExists',
)


class StreamBaseError(BaseError):
    pass


class StreamNotFound(StreamBaseError):
    message = 'Stream not found: stream_name={stream_name}'

    def __init__(self, stream_name):
        super().__init__(stream_name=stream_name)


class StreamActionNotFoundError(StreamBaseError):
    message = 'Action not found: ' \
        'stream_name={stream_name}, action_name={action_name}'

    def __init__(self, stream_name, action_name):
        super().__init__(stream_name=stream_name, action_name=action_name)


class StreamFieldNotFound(StreamBaseError):

    message = 'Field not found: ' \
        'stream_name={stream_name}, field_name={field_name}'

    def __init__(self, stream_name, field_name):
        super().__init__(stream_name=stream_name, field_name=field_name)


class StreamLimitError(StreamBaseError):

    message = 'Stream limit error ' \
        'stream_name={stream_name}, allowed_limits=1-{stream_limits}'

    def __init__(self, stream_name, stream_limits):
        super().__init__(stream_name=stream_name, stream_limits=stream_limits)


class StreamItemNotFoundError(StreamBaseError):

    message = 'Item not found ' \
        'stream_name={stream_name}, item_id={item_id}'

    def __init__(self, stream_name, item_id):
        super().__init__(stream_name=stream_name, item_id=item_id)


class StreamItemAlreadyExists(StreamBaseError):

    message = 'Item already exists ' \
        'stream_name={stream_name}, item_id={item_id}'

    def __init__(self, stream_name, item_id):
        super().__init__(stream_name=stream_name, item_id=item_id)




# -*- coding: utf-8 -*-
from io import StringIO
from datetime import datetime, timedelta
from random import randint, choice, sample, randrange, random

from iktomi.unstable.db.sqla.factories import return_locals
from ikcms.components.markup.model import ExpandableMarkup

from .vesna import phrase, randname


import logging
# logging.basicConfig()
# logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

def _create_image_file(file_manager, width, height):
    from iktomi.unstable.db.sqla.images import Image as PILImage
    image = PILImage.new('RGB', (width, height), (randint(1, 240),
                                                  randint(1, 240),
                                                  randint(1, 240), 1))
    fp = StringIO()
    image.save(fp, 'JPEG')
    fp.seek(0)
    temp_file = file_manager.create_transient(fp, '%s.%s' %
                                              (randint(1, 10000000), '.jpg'))
    fp.close()
    return temp_file
    # return None


class GeneratorField(object):
    pass


class FullName(GeneratorField):
    def __call__(self, app):
        return randname()


class FirstName(GeneratorField):
    def __call__(self, app):
        return randname().split(' ')[1]


class LastName(GeneratorField):
    def __call__(self, app):
        return randname().split(' ')[0]


class Image(GeneratorField):
    def __init__(self, width=1600, height=900):
        self.width = width
        self.height = height

    def __call__(self, app):
        file_manager = app.admin_file_manager
        return _create_image_file(file_manager, self.width, self.height)


class Date(GeneratorField):
    def __init__(self, years=None):
        self.years = years or [datetime.now().year - 1]

    def __call__(self, app):
        return datetime(choice(self.years), randint(1, 12), randint(1, 28))


class Text(GeneratorField):
    def __init__(self, min_length=20, max_length=250, html=False):
        self.min_length = min_length
        self.max_length = max_length
        self.html = html

    def __call__(self, app):
        if self.html:
            return '\n'.join(
                ['<p>%s</p>' % self.chunk() for _ in range(1, randint(2, 10))])
        return self.chunk()

    def chunk(self):
        return phrase()[:randint(self.max_length, self.max_length + 1)]


class NumberedTitle(GeneratorField):
    def __init__(self, title):
        self.title = title

    def __call__(self, app):
        return "%s %d" % (self.title, randint(1, 10000))


class EHTMLText(Text):
    def __call__(self, app):
        return ExpandableMarkup(super(EHTMLText, self).__call__(app))


class Words(GeneratorField):
    def __init__(self, min_length=4, count=1):
        self.min_length = min_length
        self.count = count

    def __call__(self, app):
        chunk = phrase()[:100]
        words = chunk.split(u' ')
        words = [word[:-1] if word.endswith(',') else word for word in words]
        words = [word.title() for word in words if len(word) > self.min_length]
        words = words[:self.count]
        return u' '.join(words)


class Slug(Words):
    pass


class Constant(GeneratorField):
    def __init__(self, value):
        self.value = value

    def __call__(self, app):
        return self.value


class Choice(GeneratorField):
    def __init__(self, choices):
        self.choices = choices

    def __call__(self, app):
        return choice(self.choices)


class TrueOrFalse(Choice):
    def __init__(self):
        self.choices = (True, False)


class Relation(GeneratorField):
    def __init__(self, model, multiple=False,
                 min_count=1, max_count=5, self_related=False):
        self.model = model
        self.multiple = multiple
        self.min_count = min_count
        self.max_count = max_count
        self.self_related = self_related

    def __call__(self, db, id=0):
        items = db.query(self.model).all()
        items = filter(lambda x: x.id != id, items)
        if self.multiple:
            max_count = min(self.max_count, len(items))
            min_count = min(self.min_count, max_count)
            return sample(items, randint(min_count, max_count))
        return choice(items)


class Generate(object):
    registered = []

    @classmethod
    def run(cls, app, names=[]):
        for f in cls.registered:
            if not names or f.name in names:
                logging.info('Generating %i item(s) for: %s', f.count, f.model)
                f(app)

    def __init__(self, model, **kwargs):
        self.model = model
        self.__dict__.update(kwargs)

    def __call__(self, f):
        def wrapper(app):
            db = app.db()
            f_locals = return_locals(f)(db, self.model)

            for n in range(0, self.count):
                plain_fields = {
                    k: v(app) for k, v in f_locals.iteritems() if
                    isinstance(v, GeneratorField) and not isinstance(v, Relation)
                }
                relation_fields = {k: v(db) for k, v in f_locals.iteritems() if
                                   isinstance(v, Relation)}
                obj = self.model(**plain_fields)
                db.add(obj)

                if hasattr(obj, '_create_versions'):
                    obj._create_versions()

                db.flush()
                [setattr(obj, k, v) for k, v in relation_fields.iteritems()]
                db.add(obj)
                if callable(getattr(obj, 'publish', None)):
                    if random() > 0.2:
                        obj.publish()

                db.commit()

        wrapper.count = self.count
        wrapper.model = self.model
        wrapper.name = f.__name__
        self.registered.append(wrapper)
        return wrapper



class Update(Generate):
    registered = []

    @classmethod
    def run(cls, app, names=[]):
        for f in cls.registered:
            if not names or f.name in names:
                logging.info('Updating item(s) for: %s', f.model)
                f(app)

    def __call__(self, f):
        def wrapper(app):
            db = app.db()
            f_locals = return_locals(f)(db, self.model)
            for obj in db.query(self.model):
                plain_fields = {
                    k: v(app) for k, v in f_locals.iteritems() if
                    isinstance(v, GeneratorField) and not isinstance(v, Relation)
                }
                relation_fields = {k: v(db, obj.id) for k, v in
                                   f_locals.iteritems() if
                                   isinstance(v, Relation)}
                for key, value in plain_fields.items():
                    setattr(obj, key, value)
                for key, value in relation_fields.items():
                    setattr(obj, key, value)
                db.commit()
                if callable(getattr(obj, 'publish', None)):
                    if random() > 0.2:
                        obj.publish()

        wrapper.model = self.model
        wrapper.name = f.__name__
        self.registered.append(wrapper)
        return wrapper


class Generator(object):

    def __init__(self):
        self.generate = type('Generate', (Generate,), dict(registered=[]))
        self.update = type('Update', (Update,), dict(registered=[]))

    def __call__(self, component, names):
        self.generate.run(component.app, names)
        self.update.run(component.app, names)




import asyncio
from unittest.mock import MagicMock

from ikcms.ws_apps.composite import App
from ikcms.ws_apps.composite.client import Client
from ikcms.ws_apps.composite import exceptions
from ikcms.ws_components.base import Component
from ikcms.utils.asynctests import asynctest

from . import base


class AppMock(base.AppMockMixin, App):

    pass


class Component1(Component):

    name = 'component1'

    def h_testhandler1(self):
        pass

    def client_init(self, env):
        self.app.log.append('c1: init')

    def client_close(self, env):
        self.app.log.append('c1: close')


class Component2(Component):

    name = 'component2'

    def h_testhandler2(self):
        pass

    def client_init(self, env):
        self.app.log.append('c2: init')

    def client_close(self, env):
        self.app.log.append('c2: close')


class AppTestCase(base.AppTestCase):

    App = App
    Client = Client

    def get_app(self, cfg=None, App=None):
        loop = asyncio.new_event_loop()
        loop = asyncio.set_event_loop(loop)
        App = App or self.App
        cfg = cfg or MagicMock()
        return App(cfg, loop=loop)

    @asynctest
    async def test_init(self):
        await super().test_init.coroutine(self)
        app = self.get_app()
        self.assertEqual(len(app.components), 0)
        self.assertEqual(app.handlers, {})

        class App(self.App):
            components = [Component1, Component2]
        app = self.get_app(App=App)
        self.assertEqual(len(app.components), 2)
        self.assertIsInstance(app.components[0], Component1)
        self.assertIsInstance(app.components[1], Component2)

        self.assertTrue(hasattr(app, 'component1'))
        self.assertTrue(hasattr(app, 'component2'))
        self.assertEqual(app.component1, app.components[0])
        self.assertEqual(app.component2, app.components[1])

        self.assertEqual(app.handlers, {
            'component1.testhandler1': app.component1.h_testhandler1,
            'component2.testhandler2': app.component2.h_testhandler2,
        })

    @asynctest
    async def test_call(self):
        await super().test_call.coroutine(self)
        class App(self.App):
            components = [Component1, Component2]
        app = self.get_app(App=App)
        app.log = []
        server = base.ServerMock([])
        await app(server, 'client')
        self.assertEqual(
            app.log,
            ['c1: init', 'c2: init', 'c1: close', 'c2: close'],
        )

    @asynctest
    async def test_call(self):
        class App(self.App):
            components = [Component1, Component2]
        app = self.get_app(App=App)
        c1 = app.get_component('component1')
        self.assertEqual(c1, app.components[0])
        c2 = app.get_component('component2')
        self.assertEqual(c2, app.components[1])

        with self.assertRaises(exceptions.ComponentNotFoundError) as ctx:
            c3 = app.get_component('component3')
        exc = ctx.exception
        self.assertEqual(exc.kwargs, {'component': 'component3'})





# -*- coding: utf-8 -*-
from jinja2 import Markup

import ikcms.components.render.base
import ikcms.components.render.jinja2


class BoundTemplate(ikcms.components.render.base.BoundTemplate):

    def render(self, template_name, **context):
        r = super(BoundTemplate, self).render(template_name, **context)
        return Markup(r)


Component = ikcms.components.render.jinja2.component(
    env_component_class=BoundTemplate
)


component = Component.create_cls



from ikcms.forms import fields


__all__ = (
    'simple_order',
    'String',
    'Int',
    'id',
    'title',
)


def simple_order(field, query, value):
    assert value in ['+', '-']
    model_field = field.context['stream'].mapper.table.c[field.name]
    if value == '+':
        return query.order_by(model_field)
    else:
        return query.order_by(model_field.desc())


class String(fields.String):
    order = simple_order


class Int(fields.Int):
    order = simple_order


class Date(fields.Date):
    order = simple_order


class id(Int):
    name = 'id'
    label = 'ID'


class title(String):
    name = 'title'
    label = 'Title'



import inspect
import logging
import os
import pwd
import sys

from ikcms.utils import DirPath
from ikcms.utils import FilePath
from ikcms.utils import cached_property


class Cfg(object):

    SITE_ID = 'base'

    @cached_property
    def ROOT_DIR(self):
        # May cause problems when running app with uWSGI,
        # need to pass ROOT explicitly (see `uwsgi_admin.py`).
        return os.path.dirname(os.path.abspath(sys.argv[0]))

    @cached_property
    def SITE_DIR(self):
        return os.path.join(self.ROOT_DIR, self.SITE_ID)

    @cached_property
    def CFG_DIR(self):
        return os.path.join(self.ROOT_DIR, 'cfg')

    @cached_property
    def TMP_DIR(self):
        return os.path.join(self.ROOT_DIR, 'tmp')

    @cached_property
    def DEFAULT_CUSTOM_CFG(self):
        return os.path.join(self.CFG_DIR, '{}.py'.format(self.SITE_ID))

    UID = 'someuid'

    LOG_LEVEL = 'INFO'
    LOG_FORMAT = '%(asctime)s: %(levelname)-5s: %(name)-15s: %(message)s'

    DOMAINS = []

    @cached_property
    def STATIC_DIR(self):
        return os.path.join(self.SITE_DIR, 'static')

    STATIC_URL = '/static/'

    def __init__(self, **kwargs):
        self._init_kwargs = kwargs
        self.update(kwargs)

    def update(self, kwargs):
        self.__dict__.update(kwargs)

    def update_from_py(self, filepath=None, silent=True):
        filepath = filepath or self.DEFAULT_CUSTOM_CFG
        assert filepath
        if silent and not os.path.isfile(filepath):
            return
        l = {}
        exec(open(filepath).read(), dict(cfg=self), l)
        self.update(l)
        for key, value in l.items():
            setattr(self, key, value)
        return self

    def config_uid(self):
        if os.getuid():
            return
        try:
            os.setgroups([])
            p = pwd.getpwnam(self.UID)
            uid = p[2]
            gid = p[3]
            os.setgid(gid)
            os.setegid(gid)
            os.setuid(uid)
            os.seteuid(uid)
        except AttributeError:
            sys.exit('UID and GID configuration variables are required ' \
                     'when is launched as root')

    def config_logging(self):
        level = logging.getLevelName(self.LOG_LEVEL)
        logging.basicConfig(
            level=level,
            format=self.LOG_FORMAT)
        logging.getLogger().setLevel(level)

    def as_dict(self):
        result = {}
        for key in dir(self):
            prop = getattr(self, key)
            if inspect.ismethod(prop):
                continue
            result[key] = prop
        return result

    def filepath(self, path, schemes=None):
        return FilePath(self, path, schemes=schemes)

    def dirpath(self, path, schemes=None):
        return DirPath(self, path, schemes=schemes)



from unittest import TestCase
from unittest import skipIf
from datetime import date
from random import randint

import sqlalchemy as sa
from iktomi.utils import cached_property

from ikcms import orm
from ikcms.orm import exc
from ikcms.utils.asynctests import asynctest
from ikcms.utils.asynctests.db import DbState
from ikcms.utils.asynctests.db import create_db_component

from tests.cfg import cfg


DB_URL1 = cfg.MYSQL_URL or cfg.POSTGRESS_URL
DB_URL2 = cfg.MYSQL_URL2 or cfg.POSTGRESS_URL2


class TestBase:

    base_mapper_cls = orm.mappers.Base
    DATABASES = {}

    items_table_keys = {'id', 'title', 'title2', 'date'}
    items_relation_keys = set()
    items_allowed_keys = items_table_keys.union(items_relation_keys)

    async def asetup(self):
        registry = orm.mappers.Registry.from_db_ids(self.DATABASES)
        self.create_mapper(registry)
        registry.create_schema()
        db = await create_db_component(self.DATABASES, registry)
        db_states = self.create_db_states(db)
        await db_states['empty'].reset(db)
        return {
            'db': db,
            'db_states': db_states,
        }

    @cached_property
    def mapper_cls(self):
        class MapperClass(self.base_mapper_cls):

            name = 'Test'

            def create_columns(self):
                return [
                    sa.Column('title', sa.String(255)),
                    sa.Column('title2', sa.String(255)),
                    sa.Column('date', sa.Date),
                ]
        return MapperClass


    def create_mapper(self, registry):
        self.mapper_cls.create(registry, db_id='admin')

    def create_db_states(self, db):
        raise NotImplementedError

    def create_test_items(self):
        return [
            {
                'id': 1,
                'title': '111t',
                'title2': '111t2',
                'date': date(2005, 4, 20),
            },
            {
                'id': 2,
                'title': '222t',
                'title2': '222t2',
                'date': date(1998, 7, 6),
            },
            {
                'id': 3,
                'title': '333t',
                'title2': '333t2',
                'date': date(2016, 12, 12),
            },
            {
                'id': 4,
                'title': '444t',
                'title2': '444t2',
                'date': date(2020, 1, 17),
            },
        ]

    def with_states(self, items, states):
        return [dict(item, state=state) for item, state in zip(items, states)]

    async def aclose(self, db, db_states):
        await db.close()

    def random_item(self, id):
        return dict(
            id=id,
            title='title {}'.format(randint(0, 1000)),
            title2='title {}'.format(randint(0, 1000)),
            date=date(randint(1990, 2020), randint(1, 12), randint(1, 28)),
        )


@skipIf(not cfg.AIO_DB_ENABLED, 'AIO DB DISABLED')
class BaseMapperTestCase(TestBase, TestCase):

    DATABASES = {'admin': DB_URL1}
    base_mapper_cls = orm.mappers.Base

    def create_db_states(self, db):
        empty_state = DbState()
        empty_state.add_table('Test', db.mappers['admin']['Test'].table)

        full_state = empty_state.copy()
        full_state['Test'].set_state(self.create_test_items())
        return {
            'empty': empty_state,
            'full': full_state,
        }

    @asynctest
    async def test_init(self, db, db_states):
        mapper = db.mappers['admin']['Test']
        self.assertEqual(mapper.registry, db.mappers)
        self.assertEqual(mapper.db_id, 'admin')
        self.assertEqual(mapper.table_keys, self.items_table_keys)
        self.assertEqual(mapper.relation_keys, self.items_relation_keys)
        self.assertEqual(mapper.allowed_keys, self.items_allowed_keys)

    @asynctest
    async def test_div_keys(self, db, db_states):
        mapper = db.mappers['admin']['Test']
        table_keys, relation_keys = mapper.div_keys()
        self.assertEqual(table_keys, self.items_table_keys)
        self.assertEqual(relation_keys, self.items_relation_keys)
        table_keys, relation_keys = mapper.div_keys(['title', 'title2'])
        self.assertEqual(table_keys, {'title', 'title2'})
        self.assertEqual(relation_keys, set())

    @asynctest
    async def test_create_table(self, db, db_states):
        self.assertEqual(
            db.mappers.metadata['admin'].tables['Test'].c.keys(),
            ['id', 'title', 'title2', 'date'],
        )
        self.assertEqual(
            db.mappers['admin']['Test'].table,
            db.mappers.metadata['admin'].tables['Test'],
        )

    @asynctest
    async def test_select(self, db, db_states):
        mapper = db.mappers['admin']['Test']
        db_state = db_states['full']
        query = mapper.query()
        async with await db() as session:
            await db_state.syncdb(session)
            item2 = db_state['Test'][2]
            item3 = db_state['Test'][3]
            item4 = db_state['Test'][4]
            # ids arg
            items = await query.id(2, 4).select_items(session)
            self.assertEqual(items, [item2, item4])
            items = await query.id(2, 4).select_items(session, keys=['title'])
            self.assertEqual(items, [
                {'id': item2['id'], 'title': item2['title']},
                {'id': item4['id'], 'title': item4['title']},
            ])

            # query arg
            _query = query.filter_by(title=item3['title'])
            items = await _query.select_items(session)
            self.assertEqual(items, [item3])

            items = await _query.select_items(session, keys=['title2'])
            self.assertEqual(
                items,
                [{'id': item3['id'], 'title2': item3['title2']}],
            )

            # ids and query args
            items = await query.id(4, 2).select_items(session)
            self.assertEqual(items, [item2, item4])

            _query = query.filter_by(title=item3['title']).id(44)
            items = await _query.select_items(session)
            self.assertEqual(items, [])
            await db_state.assert_state(self, session)

    @asynctest
    async def test_insert(self, db, db_states):
        mapper = db.mappers['admin']['Test']
        db_state = db_states['empty']

        query = mapper.query().order_by(mapper.c['id'])
        async with await db() as session:
            await db_state.syncdb(session)
            item3 = db_states['full']['Test'][3]
            item1 = db_states['full']['Test'][1]
            item4 = db_states['full']['Test'][4]

            result = await query.insert_item(session, item3)
            self.assertEqual(result, item3)
            db_state['Test'].append(item3)
            await db_state.assert_state(self, session)

            result = await query.insert_item(
                session, item1, keys=['id', 'title'])
            self.assertEqual(result, item1)
            db_state['Test'].append(dict(item1, title2=None, date=None))
            await db_state.assert_state(self, session)

            result = await query.insert_item(session, dict(item4, id=None))
            self.assertEqual(result, item4)
            db_state['Test'].append(item4)
            await db_state.assert_state(self, session)

            with self.assertRaises(exc.IntegrityError):
                await mapper.insert_item(session, item3)
            await db_state.assert_state(self, session)

    @asynctest
    async def test_update(self, db, db_states):
        mapper = db.mappers['admin']['Test']
        db_state = db_states['full']
        query = mapper.query().order_by(mapper.c['id'])
        async with await db() as session:
            await db_state.syncdb(session)
            item2 = db_state['Test'][2]
            item4 = db_state['Test'][4]
            # test without 'keys' key arg
            item = await query.update_item(
                session, 2, dict(item2, title='updated'))
            item2['title'] = 'updated'
            self.assertEqual(item, item2)
            await db_state.assert_state(self, session)

            # test with 'keys' key arg
            item = await query.update_item(
                session,
                item4['id'],
                {'title2': 'updated2'},
                keys=['title2'],
            )
            item4['title2'] = 'updated2'
            self.assertEqual(item, {'id': 4, 'title2': 'updated2'})
            await db_state.assert_state(self, session)

            query = query.filter_by(title='zzz')
            with self.assertRaises(exc.ItemNotFoundError) as e:
                await query.update_item(
                    session,
                    item4['id'],
                    {'id': 4, 'title2': 'updated2'},
                    keys=['title2'],
                )
                self.assertEqual(e.args[0], 4)
            await db_state.assert_state(self, session)

    @asynctest
    async def test_delete(self, db, db_states):
        mapper = db.mappers['admin']['Test']
        query = mapper.query().order_by(mapper.c['id'])
        db_state = db_states['full']
        async with await db() as session:
            await db_state.syncdb(session)
            # id arg
            await query.delete_item(session, 2)
            del db_state['Test'][2]
            await db_state.assert_state(self, session)
            with self.assertRaises(exc.ItemNotFoundError) as e:
                await query.delete_item(session, 2)
            self.assertEqual(e.exception.args[0], 2)
            await db_state.assert_state(self, session)

            await query.delete_item(session, 4)
            del db_state['Test'][4]
            await db_state.assert_state(self, session)

    @asynctest
    async def test_count(self, db, db_states):
        mapper = db.mappers['admin']['Test']
        query = mapper.query()
        async with await db() as session:
            await db_states['empty'].syncdb(session)
            result = await query.count_items(session)
            self.assertEqual(result, 0)

            await db_states['full'].syncdb(session)
            result = await query.count_items(session)
            self.assertEqual(result, len(db_states['full']['Test'].keys()))
            result = await query.filter_by(id=1).count_items(session)
            self.assertEqual(result, 1)
            await db_states['full'].assert_state(self, session)


@skipIf(not cfg.AIO_DB_ENABLED, 'AIO DB DISABLED')
class I18nMapperTestCase(TestBase, TestCase):

    DATABASES = {
        'admin': DB_URL1,
    }

    class base_mapper_cls(orm.mappers.I18n):
        common_keys = ['title']

    items_table_keys = {'id', 'title', 'title2', 'date', 'state'}
    items_relation_keys = set()
    items_allowed_keys = items_table_keys.union(items_relation_keys)

    def create_db_states(self, db):
        empty_state = DbState()
        empty_state.add_table('TestRu', db.mappers['admin']['ru']['Test'].table)
        empty_state.add_table('TestEn', db.mappers['admin']['en']['Test'].table)

        full_state = empty_state.copy()
        items = self.with_states(
            self.create_test_items(),
            ['normal', 'absent', 'normal', 'absent'],
        )
        full_state['TestRu'].set_state(items)
        items = self.with_states(
            self.create_test_items(),
            ['absent', 'normal', 'normal', 'absent'],
        )
        full_state['TestEn'].set_state(items)
        return {
            'empty': empty_state,
            'full': full_state,
        }

    @asynctest
    async def test_create_table(self, db, db_states):
        self.assertEqual(
            set(db.mappers.metadata['admin'].tables['TestRu'].c.keys()),
            self.items_table_keys,
        )
        self.assertEqual(
            set(db.mappers.metadata['admin'].tables['TestEn'].c.keys()),
            self.items_table_keys,
        )
        self.assertEqual(
            db.mappers['admin']['ru']['Test'].table,
            db.mappers.metadata['admin'].tables['TestRu'],
        )
        self.assertEqual(
            db.mappers['admin']['en']['Test'].table,
            db.mappers.metadata['admin'].tables['TestEn'],
        )

    @asynctest
    async def test_insert(self, db, db_states):
        mapper_ru = db.mappers['admin']['ru']['Test']
        mapper_en = db.mappers['admin']['en']['Test']
        query_ru = mapper_ru.query().order_by(mapper_ru.c['id'])
        query_en = mapper_en.query().order_by(mapper_en.c['id'])

        db_state = db_states['empty']
        test_items = self.create_test_items()
        test_item1 = test_items[0]
        test_item2 = test_items[1]
        test_item3 = test_items[2]
        test_item4 = test_items[3]

        async with await db() as session:
            await db_state.syncdb(session)
            result = await query_ru.insert_item(
                session,
                dict(test_item1, id=None),
            )
            self.assertEqual(result, dict(test_item1, state='normal'))
            db_state['TestRu'].append(dict(test_item1, state='normal'))
            db_state['TestEn'].append(
                dict(test_item1, state='absent', date=None, title2=None))
            await db_state.assert_state(self, session)

            result = await query_en.insert_item(
                session,
                dict(test_item2, id=None),
            )
            self.assertEqual(result, dict(test_item2, state='normal'))
            db_state['TestEn'].append(dict(test_item2, state='normal'))
            db_state['TestRu'].append(
                dict(test_item2, state='absent', date=None, title2=None))
            await db_state.assert_state(self, session)

            result = await query_ru.insert_item(session, test_item4)
            self.assertEqual(result, dict(test_item4, state='normal'))
            db_state['TestRu'].append(dict(test_item4, state='normal'))
            db_state['TestEn'].append(
                dict(test_item4, state='absent', date=None, title2=None))
            await db_state.assert_state(self, session)


            result = await query_en.insert_item(session, test_item3)
            self.assertEqual(result, dict(test_item3, state='normal'))
            db_state['TestEn'].append(dict(test_item3, state='normal'))
            db_state['TestRu'].append(
                dict(test_item3, state='absent', date=None, title2=None))
            await db_state.assert_state(self, session)


    @asynctest
    async def test_create_i18n_version(self, db, db_states):
        mapper_ru = db.mappers['admin']['ru']['Test']
        mapper_en = db.mappers['admin']['en']['Test']
        query_ru = mapper_ru.query().order_by(mapper_ru.c['id'])
        query_en = mapper_en.query().order_by(mapper_en.c['id'])

        db_state = db_states['empty']
        test_items = self.create_test_items()
        test_item1 = test_items[0]
        test_item2 = test_items[1]
        async with await db() as session:
            db_state = db_states['full'].copy()
            await db_state.syncdb(session)
            for id, item in db_state['TestRu'].items():
                if item['state'] == 'normal':
                    with self.assertRaises(exc.ItemNotFoundError) as e: #XXX?
                        await mapper_ru.i18n_create_version(session, id)
                    self.assertEqual(e.exception.args[0], id)
                else:
                    await mapper_ru.i18n_create_version(session, id)
                    item['state'] = 'normal'
                await db_state.assert_state(self, session)

            db_state = db_states['full'].copy()
            await db_state.syncdb(session)
            for id, item in db_state['TestEn'].items():
                if item['state'] == 'normal':
                    with self.assertRaises(exc.ItemNotFoundError) as e:
                        await mapper_en.i18n_create_version(session, id)
                    self.assertEqual(e.exception.args[0], id)
                else:
                    await mapper_en.i18n_create_version(session, id)
                    item['state'] = 'normal'
                await db_state.assert_state(self, session)


    @asynctest
    async def test_update_item(self, db, db_states):
        mapper_ru = db.mappers['admin']['ru']['Test']
        mapper_en = db.mappers['admin']['en']['Test']
        query_ru = mapper_ru.query().order_by(mapper_ru.c['id'])
        query_en = mapper_en.query().order_by(mapper_en.c['id'])

        async with await db() as session:
            db_state = db_states['full'].copy()
            await db_state.syncdb(session)
            for id, item in db_state['TestRu'].items():
                values = dict(
                    title='updated_title_{}'.format(id),
                    title2='updated_title2_{}'.format(id),
                )
                if item['state'] == 'normal':
                    await query_ru.update_item(session, id, values)
                    db_state['TestRu'][id].update(values)
                    db_state['TestEn'][id]['title'] = values['title']
                else:
                    with self.assertRaises(exc.ItemNotFoundError) as e:
                        await query_ru.update_item(session, id, values)
                    self.assertEqual(e.exception.args[0], id)
                await db_state.assert_state(self, session)

            db_state = db_states['full'].copy()
            await db_state.syncdb(session)
            for id, item in db_state['TestEn'].items():
                values = dict(
                    title='updated_title_{}'.format(id),
                    title2='updated_title2_{}'.format(id),
                )
                if item['state'] == 'normal':
                    await query_en.update_item(session, id, values=values)
                    db_state['TestEn'][id].update(values)
                    db_state['TestRu'][id]['title'] = values['title']
                else:
                    with self.assertRaises(exc.ItemNotFoundError) as e:
                        await query_en.update_item(session, id, values)
                    self.assertEqual(e.exception.args[0], id)
                await db_state.assert_state(self, session)


    @asynctest
    async def test_delete_item(self, db, db_states):
        mapper_ru = db.mappers['admin']['ru']['Test']
        mapper_en = db.mappers['admin']['en']['Test']
        query_ru = mapper_ru.query().order_by(mapper_ru.c['id'])
        query_en = mapper_en.query().order_by(mapper_en.c['id'])

        async with await db() as session:
            db_state = db_states['full'].copy()
            await db_state.syncdb(session)
            for id in list(db_state['TestRu']):
                if db_state['TestRu'][id]['state'] == 'normal':
                    await query_ru.delete_item(session, id)
                    del db_state['TestRu'][id]
                    del db_state['TestEn'][id]
                else:
                    with self.assertRaises(exc.ItemNotFoundError) as e:
                        await query_ru.delete_item(session, id)
                    self.assertEqual(e.exception.args[0], id)
                await db_state.assert_state(self, session)

            db_state = db_states['full'].copy()
            await db_state.syncdb(session)
            for id in list(db_state['TestEn']):
                if db_state['TestEn'][id]['state'] == 'normal':
                    await query_en.delete_item(session, id)
                    del db_state['TestRu'][id]
                    del db_state['TestEn'][id]
                else:
                    with self.assertRaises(exc.ItemNotFoundError) as e:
                        await query_en.delete_item(session, id)
                    self.assertEqual(e.exception.args[0], id)
                await db_state.assert_state(self, session)



@skipIf(not cfg.AIO_DB_ENABLED, 'AIO DB DISABLED')
@skipIf(not (DB_URL1 and DB_URL2), 'db url or db url2 undefined')
class PubMapperTestCase(TestBase, TestCase):

    DATABASES = {
        'admin': DB_URL1,
        'front': DB_URL2,
    }
    base_mapper_cls = orm.mappers.Pub

    items_table_keys = {'id', 'title', 'title2', 'date', 'state'}
    items_relation_keys = set()
    items_allowed_keys = items_table_keys.union(items_relation_keys)

    def create_mapper(self, registry):
        self.mapper_cls.create(registry)

    def create_db_states(self, db):
        empty_state = DbState()
        empty_state.add_table('TestAdmin', db.mappers['admin']['Test'].table)
        empty_state.add_table('TestFront', db.mappers['front']['Test'].table)

        full_state = empty_state.copy()
        items = self.with_states(
            self.create_test_items(),
            ['private', 'public', 'private', 'public'],
        )
        full_state['TestAdmin'].set_state(items)
        items = self.with_states(
            self.create_test_items(),
            ['private', 'public', 'private', 'public'],
        )
        full_state['TestFront'].set_state(items)
        return {
            'empty': empty_state,
            'full': full_state,
        }


    @asynctest
    async def test_create_table(self, db, db_states):
        self.assertEqual(
            set(db.mappers.metadata['admin'].tables['Test'].c.keys()),
            self.items_table_keys,
        )
        self.assertEqual(
            set(db.mappers.metadata['front'].tables['Test'].c.keys()),
            self.items_table_keys,
        )
        self.assertEqual(
            db.mappers['admin']['Test'].table,
            db.mappers.metadata['admin'].tables['Test'],
        )
        self.assertEqual(
            db.mappers['front']['Test'].table,
            db.mappers.metadata['front'].tables['Test'],
        )

    @asynctest
    async def test_insert_item(self, db, db_states):
        mapper1 = db.mappers['admin']['Test']
        mapper2 = db.mappers['front']['Test']
        query1 = mapper1.query().order_by(mapper1.c['id'])
        query2 = mapper2.query().order_by(mapper2.c['id'])

        db_state = db_states['empty']
        test_items = self.create_test_items()

        async with await db() as session:
            db_state = db_states['empty'].copy()
            await db_state.syncdb(session)
            for num, item in enumerate(test_items):
                id = num + 1
                result = await query1.insert_item(session, dict(item, id=None))
                self.assertEqual(result, dict(item, id=id, state='private'))
                db_state['TestAdmin'].append(dict(item, id=id, state='private'))
                db_state['TestFront'].append(dict(
                    item,
                    id=id,
                    state='private',
                    title=None,
                    title2=None,
                    date=None,
                ))
                await db_state.assert_state(self, session)

            db_state = db_states['empty'].copy()
            await db_state.syncdb(session)
            for item in test_items[::-1]:
                result = await query1.insert_item(session, item)
                self.assertEqual(result, dict(item, state='private'))
                db_state['TestAdmin'].append(dict(item, state='private'))
                db_state['TestFront'].append(dict(
                    item,
                    state='private',
                    title=None,
                    title2=None,
                    date=None,
                ))
                await db_state.assert_state(self, session)

            for item in test_items:
                with self.assertRaises(exc.IntegrityError):
                    await query1.insert_item(session, item)
                await db_state.assert_state(self, session)
                with self.assertRaises(AssertionError) as e:
                    await query2.insert_item(session, item)
                await db_state.assert_state(self, session)

    @asynctest
    async def test_delete_item(self, db, db_states):
        mapper1 = db.mappers['admin']['Test']
        mapper2 = db.mappers['front']['Test']
        query1 = mapper1.query().order_by(mapper1.c['id'])
        query2 = mapper2.query().order_by(mapper2.c['id'])

        db_state = db_states['full']

        async with await db() as session:
            await db_state.syncdb(session)
            for id in list(db_state['TestAdmin'].keys()):
                await query1.delete_item(session, id)
                del db_state['TestAdmin'][id]
                del db_state['TestFront'][id]
                await db_state.assert_state(self, session)
                with self.assertRaises(exc.ItemNotFoundError) as e:
                    await query1.delete_item(session, id)
                self.assertEqual(e.exception.args[0], id)
                await db_state.assert_state(self, session)
                with self.assertRaises(AssertionError) as e:
                    await query2.delete_item(session, id)
                await db_state.assert_state(self, session)

    @asynctest
    async def test_publish(self, db, db_states):
        mapper1 = db.mappers['admin']['Test']
        mapper2 = db.mappers['front']['Test']
        query1 = mapper1.query().order_by(mapper1.c['id'])
        query2 = mapper2.query().order_by(mapper2.c['id'])

        db_state = db_states['full']
        for front_item in db_state['TestFront'].values():
            front_item['title'] = 'different title'
            front_item['title2'] = 'different title2'
        async with await db() as session:
            await db_state.syncdb(session)
            items = zip(
                db_state['TestAdmin'].values(),
                db_state['TestFront'].values(),
            )
            for admin_item, front_item in items:
                if admin_item['state'] == 'private':
                    print(admin_item['id'], admin_item)
                    await query1.publish(session, admin_item['id'])
                    admin_item['state'] = 'public'
                    front_item.update(admin_item)
                else:
                    with self.assertRaises(AssertionError) as e:
                        await query1.publish(session, admin_item['id'])
                with self.assertRaises(AssertionError) as e:
                    await query2.publish(session, front_item['id'])
                await db_state.assert_state(self, session)


@skipIf(not cfg.AIO_DB_ENABLED, 'AIO DB DISABLED')
@skipIf(not (DB_URL1 and DB_URL2), 'db url or db url2 undefined')
class I18nPubMapperTestCase(TestBase, TestCase):

    DATABASES = {
        'admin': DB_URL1,
        'front': DB_URL2,
    }
    class base_mapper_cls(orm.mappers.I18nPub):
        common_keys = ['title']

    items_table_keys = {'id', 'title', 'title2', 'date', 'state'}
    items_relation_keys = set()
    items_allowed_keys = items_table_keys.union(items_relation_keys)

    def create_mapper(self, registry):
        self.mapper_cls.create(registry)

    def create_db_states(self, db):
        empty_state = DbState()
        empty_state.add_table(
            'TestAdminRu',
            db.mappers['admin']['ru']['Test'].table,
        )
        empty_state.add_table(
            'TestAdminEn',
            db.mappers['admin']['en']['Test'].table,
        )
        empty_state.add_table(
            'TestFrontRu',
            db.mappers['front']['ru']['Test'].table,
        )
        empty_state.add_table(
            'TestFrontEn',
            db.mappers['front']['en']['Test'].table,
        )

        full_state = empty_state.copy()
        item_states = ['absent', 'private', 'public']
        id = 0
        for admin_ru_item_state in item_states:
            for admin_en_item_state in item_states:
                for front_ru_item_state in item_states:
                    for front_en_item_state in item_states:
                        id = id + 1
                        item = dict(
                            self.random_item(id),
                            state=admin_ru_item_state,
                        )
                        full_state['TestAdminRu'].append(item)
                        item = dict(
                            self.random_item(id),
                            state=admin_en_item_state,
                        )
                        full_state['TestAdminEn'].append(item)
                        item = dict(
                            self.random_item(id),
                            state=front_ru_item_state,
                        )
                        full_state['TestFrontRu'].append(item)
                        item = dict(
                            self.random_item(id),
                            state=front_en_item_state,
                        )
                        full_state['TestFrontEn'].append(item)

        return {
            'empty': empty_state,
            'full': full_state,
        }

    @asynctest
    async def test_create_table(self, db, db_states):
        for db_id in self.mapper_cls.db_ids:
            for lang in self.mapper_cls.langs:
                metadata = db.mappers.metadata[db_id]
                table = metadata.tables['Test'+lang.capitalize()]
                self.assertEqual(
                    set(table.c.keys()),
                    self.items_table_keys,
                )
                self.assertEqual(
                    db.mappers[db_id][lang]['Test'].table,
                    table,
                )

    @asynctest
    async def test_insert_item(self, db, db_states):
        mapper_admin_ru = db.mappers['admin']['ru']['Test']
        mapper_admin_en = db.mappers['admin']['en']['Test']
        mapper_front_ru = db.mappers['front']['ru']['Test']
        mapper_front_en = db.mappers['front']['en']['Test']

        query_admin_ru = mapper_admin_ru.query()\
                                    .order_by(mapper_admin_ru.c['id'])
        query_admin_en = mapper_admin_en.query()\
                                    .order_by(mapper_admin_en.c['id'])
        query_front_ru = mapper_front_ru.query()\
                                    .order_by(mapper_front_ru.c['id'])
        query_front_en = mapper_front_en.query()\
                                    .order_by(mapper_front_en.c['id'])

        db_state = db_states['empty']
        test_items = self.create_test_items()

        async with await db() as session:
            db_state = db_states['empty'].copy()
            await db_state.syncdb(session)
            for id in (1, 2, 3):
                item = self.random_item(id)
                result = await query_admin_ru.insert_item(
                    session,
                    dict(item, id=None),
                )
                self.assertEqual(result, dict(item, state='private'))
                db_state['TestAdminRu'].append(dict(item, state='private'))
                db_state['TestFrontRu'].append(dict(
                    item,
                    state='private',
                    title=None,
                    title2=None,
                    date=None,
                ))
                db_state['TestAdminEn'].append(dict(
                    item,
                    state='absent',
                    title2=None,
                    date=None,
                ))
                db_state['TestFrontEn'].append(dict(
                    item,
                    state='absent',
                    title=None,
                    title2=None,
                    date=None,
                ))
                await db_state.assert_state(self, session)

            for id in (4, 5, 6):
                item = self.random_item(id)
                result = await query_admin_en.insert_item(
                    session,
                    dict(item, id=None),
                )
                self.assertEqual(result, dict(item, state='private'))
                db_state['TestAdminEn'].append(dict(
                    item,
                    id=id,
                    state='private',
                ))
                db_state['TestFrontEn'].append(dict(
                    item,
                    id=id,
                    state='private',
                    title=None,
                    title2=None,
                    date=None,
                ))
                db_state['TestAdminRu'].append(dict(
                    item,
                    id=id,
                    state='absent',
                    title2=None,
                    date=None,
                ))
                db_state['TestFrontRu'].append(dict(
                    item,
                    id=id,
                    state='absent',
                    title=None,
                    title2=None,
                    date=None,
                ))
                await db_state.assert_state(self, session)

            db_state = db_states['empty'].copy()
            await db_state.syncdb(session)
            for id in (3, 1, 2):
                item = self.random_item(id)
                result = await query_admin_ru.insert_item(session, item)
                self.assertEqual(result, dict(item, state='private'))
                db_state['TestAdminRu'].append(dict(item, state='private'))
                db_state['TestFrontRu'].append(dict(
                    item,
                    state='private',
                    title=None,
                    title2=None,
                    date=None,
                ))
                db_state['TestAdminEn'].append(dict(
                    item,
                    state='absent',
                    title2=None,
                    date=None,
                ))
                db_state['TestFrontEn'].append(dict(
                    item,
                    state='absent',
                    title=None,
                    title2=None,
                    date=None,
                ))
                await db_state.assert_state(self, session)

            for id in (6, 4, 5):
                item = self.random_item(id)
                result = await query_admin_en.insert_item(session, item)
                self.assertEqual(result, dict(item, state='private'))
                db_state['TestAdminEn'].append(dict(item, state='private'))
                db_state['TestFrontEn'].append(dict(
                    item,
                    state='private',
                    title=None,
                    title2=None,
                    date=None,
                ))
                db_state['TestAdminRu'].append(dict(
                    item,
                    state='absent',
                    title2=None,
                    date=None,
                ))
                db_state['TestFrontRu'].append(dict(
                    item,
                    state='absent',
                    title=None,
                    title2=None,
                    date=None,
                ))
                await db_state.assert_state(self, session)

            for item in test_items:
                for query in [query_front_ru, query_front_en]:
                    with self.assertRaises(AssertionError) as e:
                        await query.insert_item(session, item)
                    await db_state.assert_state(self, session)

                for query in [query_admin_ru, query_admin_en]:
                    with self.assertRaises(exc.IntegrityError):
                        await query.insert_item(session, item)
                    await db_state.assert_state(self, session)

    @asynctest
    async def test_delete_item(self, db, db_states):
        mapper_admin_ru = db.mappers['admin']['ru']['Test']
        mapper_admin_en = db.mappers['admin']['en']['Test']
        mapper_front_ru = db.mappers['front']['ru']['Test']
        mapper_front_en = db.mappers['front']['en']['Test']

        query_admin_ru = mapper_admin_ru.query()\
                                    .order_by(mapper_admin_ru.c['id'])
        query_admin_en = mapper_admin_en.query()\
                                    .order_by(mapper_admin_en.c['id'])
        query_front_ru = mapper_front_ru.query()\
                                    .order_by(mapper_front_ru.c['id'])
        query_front_en = mapper_front_en.query()\
                                    .order_by(mapper_front_en.c['id'])

        async with await db() as session:
            db_state = db_states['full'].copy()
            await db_state.syncdb(session)
            for id, item in list(db_state['TestAdminRu'].items()):
                if item['state'] == 'absent':
                    with self.assertRaises(exc.ItemNotFoundError) as e:
                        await query_admin_ru.delete_item(session, id)
                    self.assertEqual(e.exception.args[0], id)
                else:
                    print(item)
                    await query_admin_ru.delete_item(session, id)
                    del db_state['TestAdminRu'][id]
                    del db_state['TestAdminEn'][id]
                    del db_state['TestFrontRu'][id]
                    del db_state['TestFrontEn'][id]
                await db_state.assert_state(self, session)

                with self.assertRaises(AssertionError) as e:
                    await query_front_ru.delete_item(session, id)
                await db_state.assert_state(self, session)

            db_state = db_states['full'].copy()
            await db_state.syncdb(session)
            for id, item in list(db_state['TestAdminEn'].items()):
                if item['state'] == 'absent':
                    with self.assertRaises(exc.ItemNotFoundError) as e:
                        await query_admin_en.delete_item(session, id)
                    self.assertEqual(e.exception.args[0], id)
                else:
                    await query_admin_en.delete_item(session, id)
                    del db_state['TestAdminRu'][id]
                    del db_state['TestAdminEn'][id]
                    del db_state['TestFrontRu'][id]
                    del db_state['TestFrontEn'][id]
                await db_state.assert_state(self, session)

                with self.assertRaises(AssertionError) as e:
                    await query_front_en.delete_item(session, id)
                await db_state.assert_state(self, session)

    @asynctest
    async def test_publish(self, db, db_states):
        mapper_admin_ru = db.mappers['admin']['ru']['Test']
        mapper_admin_en = db.mappers['admin']['en']['Test']
        mapper_front_ru = db.mappers['front']['ru']['Test']
        mapper_front_en = db.mappers['front']['en']['Test']

        query_admin_ru = mapper_admin_ru.query()\
                                    .order_by(mapper_admin_ru.c['id'])
        query_admin_en = mapper_admin_en.query()\
                                    .order_by(mapper_admin_en.c['id'])
        query_front_ru = mapper_front_ru.query()\
                                    .order_by(mapper_front_ru.c['id'])
        query_front_en = mapper_front_en.query()\
                                    .order_by(mapper_front_en.c['id'])

        db_state = db_states['full']
        async with await db() as session:
            await db_state.syncdb(session)
            items = zip(
                db_state['TestAdminRu'].values(),
                db_state['TestAdminEn'].values(),
                db_state['TestFrontRu'].values(),
                db_state['TestFrontEn'].values(),
            )
            for admin_ru_item, admin_en_item, front_ru_item, front_en_item \
                    in items:
                id = admin_ru_item['id']
                if admin_ru_item['state'] == 'private':
                    await query_admin_ru.publish(session, id)
                    admin_ru_item['state'] = 'public'
                    front_ru_item.update(admin_ru_item)
                    front_en_item['title'] = admin_ru_item['title']
                elif admin_ru_item['state'] == 'absent':
                    with self.assertRaises(exc.ItemNotFoundError) as e:
                        await query_admin_ru.publish(session, id)
                    self.assertEqual(e.exception.args[0], id)
                else:
                    with self.assertRaises(AssertionError) as e:
                        await query_admin_ru.publish(session, id)
                with self.assertRaises(AssertionError) as e:
                    await query_front_ru.publish(session, id)
                await db_state.assert_state(self, session)




# coding: utf8
import six
import os
import logging
import datetime
import warnings
from collections import OrderedDict

from pytz import timezone
from babel.support import Format
import babel.dates
from babel import Locale
from jinja2 import Markup
from ikcms.utils import cached_property

from .. import base
from . import handlers
from .catalog import POCatalog
from .cli import Cli

logger = logging.getLogger(__name__)


class Lang(str):

    def __new__(cls, component, name, env=None):
        self = str.__new__(cls, name)
        self.component = component
        self.translations = component.translations[name]
        self.format = Format(name)
        self.locale = Locale(name)
        self._env = env
        return self

    def create_binded(self, env):
        return self.__class__(self.component, self, env)

    @cached_property
    def root(self):
        if self._env:
            return getattr(self._env.root, self)
        else:
            return getattr(self.component.app.root, self)

    @cached_property
    def models(self):
        return {db_id: getattr(models, self, None) \
            for db_id, models in self.component.app.db.models.items()}

    @cached_property
    def url_for(self):
        return self.root.build_url

    def months(self, type='stand-alone', form='wide'):
        return self.locale.months[type][form]

    def gettext(self, msgid):
        message = self.translations.gettext(six.u(msgid))
        if not message:
            return msgid
        if isinstance(msgid, Markup):
            message = Markup(message)
        return message

    def ngettext(self, msgid1, msgid2, n):
        message = self.translations.ngettext(
            unicode(msgid1),
            unicode(msgid2),
            n,
        )
        if not message:
            return msgid1
        if isinstance(msgid1, Markup):
            message = Markup(message)
        return message


    def _(self, msgid):
        return self.gettext(msgid)

    def format_date(self, dt, format):
        return self.component.format_date(self, dt, format)

    def format_datetime(self, dt, format):
        return self.component.format_datetime(self, dt, format)

    def date(self, dt, **kwargs):
        return self.component.date(self, dt, **kwargs)

    def datetime(self, dt, **kwargs):
        return self.component.datetime(self, dt, **kwargs)

    def daterange(self, start_dt, end_dt, **kwargs):
        return self.component.daterange(self, start_dt, end_dt, **kwargs)

    def isodate(self, dt):
        return self.component.isodate(dt)



class Component(base.Component):

    name = 'i18n'
    langs = ['ru', 'en']
    locale_dir = 'locale'
    categories = []
    lang_class = Lang

    DEFAULT_TIMEZONE = 'Europe/Moscow'

    DATE_FORMATS = {
        'ru': u"d MMMM y 'года'",
        'en': u"MMMM d, y",
    }

    DATE_FORMATS_NO_YEAR = {
        'ru': u"d MMMM",
        'en': u"MMMM d",
    }

    DATETIME_FORMATS = {
        'ru': u"d MMMM y 'года', HH:mm",
        'en': u"MMMM d, y, HH:mm",
    }

    DATETIME_FORMATS_NO_YEAR = {
        'ru': u"d MMMM, HH:mm",
        'en': u"MMMM d, HH:mm",
    }

    DATERANGE_FORMATS = {
        'ru': {
            # y1=?y2, m1=?m2, d1=?d2 : (format1, format2)
            ''   : (u'd MMMM y \N{EN DASH} ', u"d MMMM y 'года'"),
            'y'  : (u'd MMMM \N{EN DASH} ', u"d MMMM y 'года'"),
            'ym' : (u'd\N{EN DASH}', u"d MMMM y 'года'"),
            'ymd': (u"d MMMM y 'года'", u''),
        },
        'en': {
            ''   : (u'MMMM d, y \N{EN DASH} ', u'MMMM d, y'),
            'y'  : (u'MMMM d \N{EN DASH} ', u'MMMM d, y'),
            'ym' : (u'MMMM d\N{EN DASH}', u'd, y'),
            'ymd': (u'MMMM d, y', u''),
        },
    }

    method_map = [
        ('**.py', 'python'),
        ('**.html', 'jinja2.ext:babel_extract'),
    ]

    options_map = {
        '**.html': {},
        '**.py': {},
    }

    @cached_property
    def catalogs(self):
        site_id = self.app.cfg.SITE_ID
        return [
            {
                'name': site_id,
                'input-dirs': [site_id],
                'pot-file': 'locale/{}.pot'.format(site_id),
                'po-files': {lang: 'locale/{}/{}.po'.format(lang, site_id) \
                    for lang in self.langs},
            },
            {
                'name': 'ikcms',
                'po-files': {
                    'ru': 'pkg://ikcms/locale/ru/messages.po',
                    'en': 'pkg://ikcms/locale/en/messages.po',
                },
            },
        ]


    def __init__(self, app):
        super(Component, self).__init__(app)
        tz_str = getattr(app.cfg, 'TIMEZONE', self.DEFAULT_TIMEZONE)
        self.timezone = timezone(tz_str)
        self._create_catalogs()
        self._create_translations()
        self._create_langs()

    def set_lang(self, env, lang):
        assert lang in self.langs
        env.langs = {lang: lang.create_binded(env) \
            for lang in self.langs.values()}
        env.lang = env.langs[lang]

    def h_lang(self, lang):
        return handlers.HLang(self, lang)

    def h_for_langs(self, *langs):
        return handlers.HForLangs(self, *langs)

    def format_date(self, locale, dt, format):
        return babel.dates.format_date(dt, format, locale=locale)

    def format_datetime(self, locale, dt, format):
        return babel.dates.format_datetime(dt, format, locale=locale)

    def date(self, locale, dt, **kwargs):
        optional_year = kwargs.get('optional_year', False)
        format = self.DATE_FORMATS[locale]
        if optional_year:
            today = datetime.date.today()
            if today.year == dt.year and (today.month, dt.month) != (12, 1):
                format = self.DATE_FORMATS_NO_YEAR[locale]
        return babel.dates.format_date(dt, format, locale=locale)

    def datetime(self, locale, dt, **kwargs):
        optional_year = kwargs.get('optional_year', False)
        format = self.DATETIME_FORMATS[locale]
        relative = kwargs.get('relative', False)
        if relative:
            delta = datetime.datetime.now() - dt
            if not delta.days:
                return babel.dates.format_timedelta(
                    -delta,
                    locale=locale,
                    add_direction=True,
                )
        if optional_year:
            today = datetime.date.today()
            if today.year == dt.year and (today.month, dt.month) != (12, 1):
                format = self.DATETIME_FORMATS_NO_YEAR[locale]
        return babel.dates.format_datetime(dt, format, locale=locale)

    def daterange(self, locale, start_dt, end_dt, **kwargs):
        if end_dt is None:
            end_dt = start_dt
        key = ''
        for attr in ['year', 'month', 'day']:
            if getattr(start_dt, attr)!=getattr(end_dt, attr):
                break
            key += attr[0]
        start_format, end_format = self.RANGE_FORMATS[locale][key]
        return self.format_date(locale, start_dt, start_format) + \
            self.format_date(lcoale, end_dt, end_format)

    def isodate(self, dt):
        return self.timezone.localize(dt).isoformat()

    def _create_catalogs(self):
        catalogs_cfg = self.catalogs
        self.catalogs = OrderedDict()
        for catalog_cfg in catalogs_cfg:
            name = catalog_cfg['name']
            po_files = {lang: self.app.cfg.filepath(path) \
                    for lang, path in catalog_cfg['po-files'].items()}
            input_dirs = [self.app.cfg.dirpath(dirpath, ['']) \
                for dirpath in catalog_cfg.get('input-dirs', [])]
            pot_file = catalog_cfg.get('pot-file')
            if pot_file:
                pot_file = self.app.cfg.filepath(pot_file, [''])

            catalog = POCatalog(
                name,
                po_files=po_files,
                input_dirs=input_dirs,
                pot_file=pot_file,
                method_map=self.method_map,
                options_map=self.options_map,
            )
            self.catalogs[name] = catalog

    def _create_translations(self):
        self.translations = {}
        for lang in self.langs:
            translations = POCatalog.Translations(None, lang)
            for name, catalog in self.catalogs.items():
                catalog_translations = catalog.get_translations(lang)
                if catalog_translations:
                    translations.add(catalog_translations)
            self.translations[lang] = translations

    def _create_langs(self):
        self.langs = {lang: self.lang_class(self, lang) for lang in self.langs}

component = Component.create_cls





import collections
from unittest.mock import MagicMock

import sqlalchemy as sa
import ikcms.ws_components.db


class TableState(dict):

    def __init__(self, table, primary_keys=None):
        super().__init__()
        self.table = table
        self.primary_keys = primary_keys or ['id']
        self.order_fields = [table.c[key] for key in self.primary_keys]

    def row_key(self, row):
        if len(self.primary_keys) == 1:
            return row[self.primary_keys[0]]
        else:
            return tuple([row[key] for key in self.primary_keys])

    async def set_db_state(self, session, rows=None):
        if rows is None:
            rows = self.get_state()
        else:
            self.set_state(rows)
        await session.execute(sa.sql.delete(self.table))
        if rows:
            await session.execute(sa.sql.insert(self.table).values(rows))
    syncdb = set_db_state

    async def get_db_state(self, session):
        q = sa.sql.select(self.table.c).order_by(*self.order_fields)
        result = await session.execute(q)
        return [dict(row) for row in result]

    def get_state(self):
        return [self[key] for key in sorted(self.keys())]

    def set_state(self, rows):
        self.clear()
        for row in rows:
            self.append(row.copy())

    def append(self, row):
        self[self.row_key(row)] = row

    async def assert_state(self, test, session):
        db_state = await self.get_db_state(session)
        state = self.get_state()
        test.assertEqual(db_state, state, self.table.name)

    def copy(self):
        state = self.__class__(self.table, self.primary_keys)
        state.set_state(self.get_state())
        return state


class DbState(collections.OrderedDict):

    def copy(self):
        return self.__class__(
            [(key, value.copy()) for key, value in self.items()])

    def add_table(self, key, table, primary_keys=None):
        self[key] = TableState(table, primary_keys or ['id'])

    async def syncdb(self, session):
        for table_state in self.values():
            await table_state.set_db_state(session)

    async def assert_state(self, test, session):
        for table_state in self.values():
            await table_state.assert_state(test, session)


    async def reset(self, db):
        async with await db() as session:
            tables = [state.table for state in self.values()]
            for table in tables[::-1]:
                conn = await session.get_connection(db.binds[table])
                try:
                    await conn.execute(sa.sql.ddl.DropTable(table))
                except Exception:
                    pass
            for table in tables:
                conn = await session.get_connection(db.binds[table])
                await conn.execute(sa.sql.ddl.CreateTable(table))


async def create_db_component(DATABASES, registry):
    app = MagicMock()
    del app.db
    app.cfg.DATABASES = DATABASES

    Component = ikcms.ws_components.db.component(mappers=registry)
    return await Component.create(app)


def assert_client_error(test, ctx, test_exc):
    exc = ctx.exception
    client_error = ClientError(test_exc)
    test.assertEqual(exc.error, client_error.error)
    test.assertEqual(exc.kwargs, client_error.kwargs)



from iktomi.utils import cached_property
import ikcms.cfg.base


class Cfg(ikcms.cfg.base.Cfg):

    SITE_ID = 'tests'
    MYSQL_URL = 'mysql://root@localhost/test?charset=utf8'
    POSTGRESS_URL = 'postgress://root@localhost/test?charset=utf8'
    MYSQL_URL2 = 'mysql://root@localhost/test2?charset=utf8'
    POSTGRESS_URL2 = 'postgress://root@localhost/test2?charset=utf8'

    MEMCACHE_HOST = 'localhost'
    MEMCACHE_PORT = 11211

    @cached_property
    def AIO_DB_ENABLED(self):
        try:
            import sqlalchemy
            return self.AIOMYSQL_ENABLED or self.AIOPG_ENABLED
        except ImportError:
            return False

    @cached_property
    def AIOMYSQL_ENABLED(self):
        try:
            import aiomysql
            return bool(self.MYSQL_URL)
        except ImportError:
            return False

    @cached_property
    def AIOPG_ENABLED(self):
        return False #XXX
        try:
            import aiopg
            return bool(self.POSTGRESS_URL)
        except ImportError:
            return False

    @cached_property
    def DB_URL(self):
        return self.MYSQL_URL or self.POSTGRESS_URL

    @cached_property
    def JINJA2_ENABLED(self):
        try:
            import jinja2
            return True
        except ImportError:
            return False

    @cached_property
    def AIOMCACHE_ENABLED(self):
        try:
            import aiomcache
            return True
        except ImportError:
            return False

    @cached_property
    def AIOREDIS_ENABLED(self):
        try:
            import aioredis
            return True
        except ImportError:
            return False





cfg = Cfg()
cfg.update_from_py()



from code import interact
from ikcms.ws_servers.websockets import WS_Server
from .base import Cli


class WsAppCli(Cli):

    name = 'app'
    server_class = WS_Server

    def command_serve(self, level=None, cfg=''):
        kwargs = dict(custom_cfg_path=cfg)
        if level:
            kwargs['LOG_LEVEL'] = level
        app = self.create_app(**kwargs)
        server = self.server_class(app.cfg.WS_SERVER_URL, app)
        server.serve_forever()

    def command_shell(self, level=None, cfg=''):
        kwargs = dict(custom_cfg_path=cfg)
        if level:
            kwargs['LOG_LEVEL'] = level
        app = self.create_app(**kwargs)
        namespace = self.shell_namespace(app)
        interact('Namespace {!r}'.format(namespace),
                 local=namespace)

    def shell_namespace(self, app):
        return {
            'app': app,
        }




import sys

from ikcms.cli.base import Cli


class Cli(Cli):

    name = 'i18n'

    def command_list(self):
        """ List available catalogs """
        app = self.create_app()
        print('\n'.join(app.i18n.catalogs.keys()))

    def command_extract(self, *names):
        """ Extract translations in po file """
        app = self.create_app()
        for catalog in self._get_catalogs(app, names):
            catalog.extract()

    def command_merge(self, *names):
        """ Extract translations in po file """
        app = self.create_app()
        for catalog in self._get_catalogs(app, names):
            catalog.merge()

    def _get_catalogs(self, app, names):
        for name in names:
            if name not in app.i18n.catalogs:
                print('Unknown catalog {}'.format(name))
                sys.exit()
            if app.i18n.catalogs[name].read_only:
                print('Catalog {} is read only'.format(name))
                sys.exit()
        if not names:
            names = list(app.i18n.catalogs)
        return [app.i18n.catalogs[name] for name in names]




from collections import OrderedDict

from . import exceptions
from . import convs


class Form(OrderedDict):

    fields = []

    def __init__(self, **context):
        super().__init__()
        context.setdefault('form', self)
        self.conv = convs.RawDict(self)
        for field in self.fields:
            assert field.name
            self[field.name] = field(context)

    def list(self, keys=None):
        if keys is not None:
            assert not set(keys) - set(self)
            return [(key, value) for key, value in self.items() if key in keys]
        else:
            return self.items()

    def to_python(self, raw_values, keys=None):
        raw_dict = self.conv.to_python(raw_values)
        python_dict = {}
        errors = {}
        for name, field in self.list(keys):
            try:
                python_dict.update(field.to_python(raw_dict))
            except exceptions.ValidationError as exc:
                errors.update(exc.kwargs['error'])
        return python_dict, errors

    def from_python(self, python_values, keys=None):
        python_dict = self.conv.to_python(python_values)
        raw_dict = {}
        for name, field in self.list(keys):
            raw_dict.update(field.from_python(python_dict))
        return raw_dict

    def values_to_python(self, raw_values):
        return [self.to_python(raw_value) for raw_value in raw_values]

    def values_from_python(self, values):
        return [self.from_python(value) for value in values]

    def get_cfg(self):
        return [f.widget.to_dict(f) for f in self.values()]

    def get_initials(self, **kwargs):
        values = {}
        for name, field in self.items():
            values[name] = field.get_initials(**kwargs)
        return values





import sqlalchemy as sa

class Query(sa.sql.Select):

    def __init__(self, mapper, *args, **kwargs):
        self.mapper = self.m = mapper
        super().__init__([self.mapper.c['id']], *args, **kwargs)

    def id(self, *ids):
        if len(ids) == 1:
            return self.where(self.mapper.c['id'] == ids[0])
        else:
            return self.where(self.mapper.c['id'].in_(ids))

    def filter_by(self, **kwargs):
        q = self
        for key, value in kwargs.items():
            q = q.where(self.mapper.c[key] == value)
        return q

    async def select_items(self, session, keys=None):
        return await self.mapper.select_items(
            session,
            self,
            keys=keys,
        )

    async def select_first_item(self, session, keys=None):
        return await self.mapper.select_first_item(
            session,
            self,
            keys=keys,
        )

    async def insert_item(self, session, values, keys=None): #XXX
        return await self.mapper.insert_item(
            session,
            values=values,
            keys=keys,
        )

    async def update_item(self, session, item_id, values, keys=None):
        return await self.mapper.update_item(
            session,
            self,
            item_id=item_id,
            values=values,
            keys=keys,
        )

    async def delete_item(self, session, item_id):
        return await self.mapper.delete_item(
            session,
            self,
            item_id,
        )

    async def count_items(self, session):
        return await self.mapper.count_items(
            session,
            self,
        )

    async def fill(self, session, data, path):
        return await self.mapper.fill(
            session,
            self,
            data,
            path,
        )



class PubQuery(Query):

    async def publish(self, session, item_id):
        return await self.mapper.publish(session, self, item_id)





import six
import os
import warnings
from pkg_resources import resource_stream
import logging

from gettext import c2py

from babel import Locale
from babel.support import Translations
from babel.messages.plurals import get_plural
from babel.messages.catalog import Catalog
from babel.messages.extract import extract_from_dir
from babel.messages.pofile import read_po
from babel.messages.pofile import write_po

logger = logging.getLogger(__name__)


class POTranslations(Translations):

    def __init__(self, fp, locale):
        if not isinstance(locale, Locale):
            locale = Locale.parse(locale)
        self.locale = locale
        super(POTranslations, self).__init__(fp)
        self.plural = c2py(get_plural(locale).plural_expr)

    def _parse(self, fp):
        catalog = read_po(fp, locale=self.locale)
        self._catalog = c = {}
        for message in six.itervalues(catalog._messages):
            if message.pluralizable:
                for idx, string in enumerate(message.string):
                    c[message.id[0], idx] = string
            else:
                c[message.id] = message.string

    def ngettext(self, *args):
        return Translations.ngettext(self, *args)



class POCatalog(object):

    Translations = POTranslations

    def __init__(
            self,
            name,
            po_files,
            input_dirs=[],
            pot_file=None,
            method_map=[],
            options_map={},
            root_dir=None
        ):
        assert not (bool(input_dirs) ^ bool(pot_file))
        self.name = name
        self.po_files = po_files.copy()
        self.input_dirs = list(input_dirs)
        self.pot_file = pot_file
        self.method_map = method_map
        self.options_map = options_map
        self.root_dir = root_dir

    def extract(self):
        if not self.input_dirs:
            return
        logger.info('Processing catalog "%s"', self.name)
        catalog = Catalog()
        for dirpath in self.input_dirs:
            extracted = extract_from_dir(
                str(dirpath),
                self.method_map,
                self.options_map,
            )
            for filename, lineno, message, comments, context in extracted:
                fpath = dirpath.join(filename)
                if self.root_dir:
                    fpath = os.path.relpath(fpath, self.root_dir)
                logger.info('Extracting messages from %s', fpath)
                catalog.add(
                    message,
                    None,
                    [(fpath, lineno)],
                    auto_comments=comments,
                    context=context,
                )

        logger.info('Writing PO template file to %s', str(self.pot_file))
        self.pot_file.makedirs()
        with self.pot_file.open('w') as pot_fp:
            write_po(pot_fp, catalog, omit_header=True)

    def merge(self):
        if not self.input_dirs:
            return
        logger.info('Processing catalog "%s"', self.name)
        if not self.pot_file.exists():
            logger.warning(
                "POT template file %s doesn't exist",
                str(self.pot_file),
            )
            return
        for lang, po_file in self.po_files.items():
            if po_file.isreadonly():
                raise ValueError('File is readonly', str(po_file))
            with self.pot_file.open() as pot_fp:
                template = read_po(pot_fp, locale=lang)
            if po_file.exists():
                logger.info('Merging template to PO file %s', po_file)
                with po_file.open('U') as po_fp:
                    catalog = read_po(po_fp, locale=lang)
                catalog.update(template)
            else:
                logger.info('Creating new PO file %s', po_file)
                catalog = Catalog()
                catalog = template
                catalog.locale = Locale.parse(lang)
                catalog.fuzzy = False
            tmp_file = str(po_file) + '.new'
            po_file.makedirs()
            with open(tmp_file, 'wb') as tmp_fp:
                write_po(tmp_fp, catalog, omit_header=True)
            os.rename(tmp_file, str(po_file))

    def get_translations(self, lang):
        po_file = self.po_files.get(lang)
        if not po_file:
            return None
        if not po_file.exists():
            warnings.warn("File {} doesn't exist".format(po_file))
            return None
        with po_file.open('U') as po_fp:
            translations = POTranslations(po_fp, lang)
        return translations





import time
import random
import logging
import pickle

import sqlalchemy as sa


logger = logging.getLogger(__name__)


class CachedModel(object):

    check_timeout = 5
    lock_timeout = 5
    checked_ts = 0
    updated_ts = 0
    created_ts = 0
    prefix = None

    def __init__(self, component, model_path, prefix=None):
        self.component = component
        self.app = component.app
        self.model_path = model_path
        self.model = self.app.db.get_model(model_path)
        self.preview = hasattr(self.app, 'preview')
        self.front_db_id = self.preview and 'admin' or 'front'
        if prefix:
            self.prefix = prefix
        self.init_cache()

    def reset_cache(self):
        tmp = self.__class__(
            self.component,
            self.model_path,
            prefix=str(random.randint(1000, 100000)),
        )

        with self.app.cache.pipe() as pipe:
            keys = [
                'checked_ts',
                'updated_ts',
                'created_ts',
                'items',
                'indexes',
            ]
            for key in keys:
                if self.app.cache.client.exists(tmp._cache_key(key)):
                    pipe.rename(
                        tmp._cache_key(key),
                        self._cache_key(key),
                    )
            pipe.execute()



    def init_cache(self):
        if self.app.cache.get(self._cache_key('created_ts')):
            return
        with self.app.cache.lock(
            self._cache_key('lock'),
            expires=self.lock_timeout,
        ) as lock:
            self.update_cache()

    def update_cache(self):
        if time.time() < (self.checked_ts + self.check_timeout):
            return

        # check update timeout
        cache_checked_ts, cache_updated_ts, cache_created_ts = \
            self.app.cache.mget(
                self._cache_key('checked_ts'),
                self._cache_key('updated_ts'),
                self._cache_key('created_ts'),
            )
        try:
            self.checked_ts = cache_checked_ts = int(cache_checked_ts)
            self.updated_ts = cache_updated_ts = int(cache_updated_ts)
            self.created_ts = cache_created_ts = int(cache_created_ts)
        except (TypeError, ValueError):
            cache_checked_ts = None
            cache_updated_ts = None

        # if check timeout not expired, updating is not required
        if cache_checked_ts:
            if time.time() < (cache_checked_ts + self.check_timeout):
                return cache_updated_ts

        # if other process is already updating cache, we do nothing
        if not self.app.cache.add(
                self._cache_key('updating'),
                1,
                self.lock_timeout,
        ):
            return cache_updated_ts

        now_ts = int(time.time())
        try:
            db_updated_ts = self.get_updated_ts_from_db()
            if db_updated_ts is None:
                return None
        except sa.exc.DBAPIError as exc:
            logger.warning('Retrieve {} error: {}'.format(self.model, exc))
            return None

        # if db not changed, we update cache checked ts
        if cache_updated_ts and db_updated_ts <= cache_updated_ts:
            with self.app.cache.pipe() as pipe:
                pipe.set(self._cache_key('checked_ts'), now_ts)
                pipe.delete(self._cache_key('updating'))
                pipe.execute()
                self.checked_ts = now_ts
            return cache_updated_ts

        # Get items from db
        try:
            items = self.get_items_from_db()
        except sa.exc.DBAPIError as exc:
            logger.warning('Retrieve sections error: {}'.format(exc))
            return None

        indexes = self.create_indexes(items)
        raw_items = {id: self._dumps(item) for id, item in items.items()}
        raw_indexes = {id: self._dumps(item) for id, item in indexes.items()}

        # Update cache
        with self.app.cache.pipe() as pipe:
            pipe.set(self._cache_key('updated_ts'), db_updated_ts)
            pipe.set(self._cache_key('checked_ts'), now_ts)
            pipe.set(self._cache_key('created_ts'), now_ts)
            if raw_items:
                pipe.hmset(self._cache_key('items'), raw_items)
            if raw_indexes:
                pipe.hmset(self._cache_key('indexes'), raw_indexes)
            pipe.delete(self._cache_key('updating'))
            pipe.execute()
        self.updated_ts = db_updated_ts
        logging.info('{} cache updated: {}'.format(
            self.model_path,
            time.time()-now_ts,
        ))

        return db_updated_ts

    def version(self, session):
        return self.app.cache.get(self._cache_key('created_ts'))

    def get_updated_ts_from_db(self):
        session = self.app.db()
        s = sa.sql.select([sa.func.max(self.model.updated_dt)])
        result = list(session.execute(s))
        session.close()
        if result and result[0][0]:
            return int(time.mktime(result[0][0].timetuple()))
        else:
            return None

    def get_items_from_db(self):
        session = self.app.db()
        db_objs = self._get_objs_from_db(session)
        items = {obj.id: obj.to_dict() for obj in db_objs}
        session.close()
        return items

    def get_items(self, ids):
        if not ids:
            return []
        raw_items = self.app.cache.hmget(self._cache_key('items'), ids)
        items = []
        for id, item in zip(ids, raw_items):
            if item is not None:
                item = self._loads(item)
            items.append(item)
        return items

    def get_all(self):
        raw_items = self.app.cache.hvals(self._cache_key('items'))
        return [self._loads(item) for item in raw_items]

    def get(self, id, default=None):
        item = self.get_items([id])[0]
        if item is None:
            return default
        else:
            return item

    def get_index(self, name, default=Exception):
        raw_index = self.app.cache.hget(self._cache_key('indexes'), name)
        if not raw_index:
            if default is Exception:
                raise Exception('Index "{}" not found'.format(name))
            else:
                return default
        return self._loads(raw_index)

    def create_indexes(self, items):
        return {}

    def _get_objs_from_db(self, session):
        return session.query(self.model).all()


    def _cache_key(self, name):
        return '{}{}:{}'.format(self.prefix or '', self.model_path, name)


    def _query(self, session):
        return session.query(self.model)

    def _dumps(self, obj):
        return pickle.dumps(obj)

    def _loads(self, string):
        return pickle.loads(string)

    def __iter__(self):
        return self.get_all().__iter__()




from datetime import date, datetime

from . import exceptions


__all__ = [
    'Converter',
    'RawDict',
    'RawList',
    'Dict',
    'List',
    'Str',
    'Int',
    'IntStr',
    'Bool',
    'Date',
]


class Converter:
    raw_type = None
    python_type = None

    def __init__(self, field):
        self.field = field

    def to_python(self, raw_value):
        if raw_value is None:
            if self.field.not_none:
                raise exceptions.RawValueNoneNotAllowedError(self.field.name)
            return None
        if self.raw_type:
            if isinstance(raw_value, self.raw_type):
                return raw_value
            else:
                raise exceptions.RawValueTypeError(
                    self.field.name,
                    self.raw_type,
                )
        else:
            return raw_value

    def from_python(self, python_value):
        if python_value is None:
            if self.field.not_none:
                raise exceptions.PythonValueNoneNotAllowedError(self.field.name)
            return None
        if self.python_type:
            if isinstance(python_value, self.python_type):
                return python_value
            else:
                raise exceptions.PythonValueTypeError(
                    self.field.name,
                    self.python_type,
                )
        else:
            return python_value




class RawDict(Converter):
    raw_type = dict
    python_type = dict


class RawList(Converter):
    raw_type = list
    python_type = list


class Str(Converter):
    raw_type = str
    python_type = str


class Int(Converter):
    raw_type = int
    python_type = int


class IntStr(Converter):
    raw_type = str
    python_type = int
    error_not_valid = 'Not a valid integer'

    def to_python(self, raw_value):
        raw_value = super().to_python(raw_value)
        if raw_value is None:
            return None
        try:
            return int(raw_value)
        except ValueError:
            raise exceptions.ValidationError(self.error_not_valid)

    def from_python(self, value):
        value = super().from_python(value)
        if value is None:
            return value
        return str(value)


class Bool(Converter):
    raw_type = bool
    python_type = bool


class Dict(Converter):
    raw_type = dict
    python_type = dict

    def to_python(self, raw_dict):
        raw_dict = super().to_python(raw_dict)
        if raw_dict is None:
            return None
        python_dict = {}
        errors = {}
        for subfield in self.field.fields:
            try:
                python_dict.update(subfield.to_python(raw_dict))
            except exceptions.ValidationError as exc:
                errors.update(exc.error)
        if errors:
            raise exceptions.ValidationError(errors)
        return python_dict

    def from_python(self, python_dict):
        python_dict = super().from_python(python_dict)
        if python_dict is None:
            return None
        raw_dict = {}
        for subfield in self.field.fields:
            raw_dict.update(subfield.from_python(python_dict))
        return raw_dict


class List(Converter):
    raw_type = list
    python_type = list

    def __init__(self, field):
        super().__init__(field)
        assert len(field.fields)
        assert field.fields[0].name is None
        self.item_field = field.fields[0]

    def to_python(self, raw_list):
        raw_list = super().to_python(raw_list)
        if raw_list is None:
            return None
        python_list = []
        errors = []
        for raw_value in raw_list:
            try:
                python_value = self.item_field.to_python({None: raw_value})[None]
                python_list.append(python_value)
            except exceptions.ValidationError as exc:
                errors.append(exc.error[None])
            else:
                errors.append(None)
        if any(errors):
            raise exceptions.ValidationError(errors)
        return python_list

    def from_python(self, python_list):
        python_list = super().from_python(python_list)
        if python_list is None:
            return None
        return [self.item_field.from_python({None: value})[None] \
                for value in python_list]


class Date(Converter):
    raw_type = str
    python_type = date
    error_not_valid = 'Not a valid date'

    def to_python(self, raw_value):
        value = super().to_python(raw_value)
        if value is None:
            return None
        try:
            return datetime.strptime(raw_value, self.field.format).date()
        except ValueError:
            raise exceptions.ValidationError(self.error_not_valid)

    def from_python(self, python_value):
        python_value = super().from_python(python_value)
        if python_value is None:
            return None
        return python_value.strftime(self.field.format)






from iktomi.utils import cached_property

from .forms import Form
from . import actions
from . import exceptions


__all__ = (
    'Base',
    'I18nMixin',
    'PublicationMixin',
    'Stream',
    'I18nStream',
    'PubI18nStream',
)


class Base:
    name = None
    title = u'Название потока'
    actions = []

    def __init__(self, component):
        assert self.name is not None
        self.component = component
        self.streams = component.streams
        self.actions = [action(self) for action in self.actions]
        self.id = self.get_id()

    @classmethod
    def create(cls, component, registry, **kwargs):
        assert cls.name is not None
        stream = cls(component, **kwargs)
        registry[stream.id] = stream

    def get_id(self, **kwargs):
        return kwargs.get('name', self.name)

    def h_action(self, env, message):
        action_name = message.get('action')
        action = self.get_action(action_name)
        if action:
            return action.handle(env, message)
        else:
            raise exceptions.StreamActionNotFoundError(self, action_name)

    def get_action(self, name):
        for action in self.actions:
            if action.name == name:
                return action

    def get_cfg(self, env):
        return {
            'name': self.name,
            'title': self.title,
            'actions': [action.get_cfg(env) for action in  self.actions],
        }


class Stream(Base):
    mapper_name = None
    db_id = 'main'

    widget = 'Stream'
    max_limit = 100

    ListForm = Form
    FilterForm = Form
    ItemForm = Form

    list_fields = []
    filter_fields = []
    item_fields = []

    permissions = {
        'streams.read': 'rx',
        'streams.edit': 'rxwcd',
    }

    default_order = ['+id']

    actions = [
        actions.List,
        actions.GetItem,
        actions.NewItem,
        actions.CreateItem,
        actions.UpdateItem,
        actions.DeleteItem,
        #actions.create_draft,
        #actions.delete_draft,
        #actions.create_new_item,
        #actions.edit_item,
        #actions.get_field_block,
        #actions.set_field_value,
        #actions.delete_item,
    ]

    def __init__(self, component):
        super().__init__(component)
        assert self.mapper_name

    @cached_property
    def mapper(self):
        mappers = self.component.app.db.mappers
        return mappers[self.db_id][self.mapper_name]

    def get_list_form(self, env):
        class ListForm(self.ListForm):
            fields = self.list_fields
        return ListForm(env=env, stream=self)

    def get_filter_form(self, env):
        class FilterForm(self.FilterForm):
            fields = self.filter_fields
        return FilterForm(env=env, stream=self)

    def get_order_form(self, env):
        class ListForm(self.ListForm):
            fields = [f for f in self.list_fields if f.order]
        return ListForm(env=env, stream=self)

    def get_item_form(self, env, item=None, kwargs=None):
        kwargs = kwargs or {}
        class ItemForm(self.ItemForm):
            fields = self.item_fields
        return ItemForm(env=env, stream=self)

    def query(self):
        return self.mapper.query()

    def get_cfg(self, env):
        list_form = self.get_list_form(env)
        filter_form = self.get_filter_form(env)
        return dict(
            super().get_cfg(env),
            widget=self.widget,
            max_limit=self.max_limit,
            list_fields=list_form.get_cfg(),
            filter_fields=filter_form.get_cfg(),
            permissions=list(self.component.app.auth.get_user_perms(
                env.user,
                self.permissions,
            )),
        )

    async def get_item(self, env, session, item_id, keys=None):
        return await self.query().id(item_id).select_first_item(session, keys)

    async def list_items(
            self,
            env,
            session,
            filters=None,
            order=None,
            page=None,
            page_size=None,
            keys=None,
    ):

        query = self.query()
        query = self._filter_query(env, query, filters)
        query = self._order_query(env, query, order)
        query = self._page_query(env, query, page, page_size)
        return await query.select_items(session, keys=keys)

    async def count_items(self, env, session, filters=None):
        query = self.query()
        query = self._filter_query(env, query, filters)
        return await query.count_items(session)

    async def new_item(self, env, kwargs):
        item_fields_form = self.get_item_form(env, kwargs=kwargs)
        return item_fields_form.get_initials(**kwargs)

    async def insert_item(self, env, session, item):
        if 'id' in item:
            if await self.is_item_exists(env, session, item['id']):
                raise exceptions.StreamItemAlreadyExistsError(
                    self.name,
                    item['id'],
                )
        return await self.query().insert_item(session, item)

    async def update_item(self, env, session, item_id, values):
        keys = list(values.keys())
        if not await self.is_item_exists(env, session, item_id):
            raise exceptions.StreamItemNotFoundError(self.name, item_id)
        return await self.query().update_item(session, item_id, values, keys)

    async def delete_item(self, env, session, item_id):
        if not await self.is_item_exists(env, session, item_id):
            raise exceptions.StreamItemNotFoundError(self.name, item_id)
        return await self.query().delete_item(session, item_id)

    async def check_perms(self, env, perms):
       return self.component.app.auth.check_perms(env.user, perms)

    async def is_item_exists(self, env, session, item_id):
        return bool(await self.get_item(env, session, item_id, ['id']))

    def _filter_query(self, env, query, filters=None):
        filters = filters or {}
        filter_form = self.get_filter_form(env)
        for name, field in filter_form.items():
            query = field.filter(query, filters.get(name))
        return query

    def _order_query(self, env, query, order=None):
        order = order or ['+id']
        order_form = self.get_order_form(env)
        for value in order:
            value, name = value[0], value[1:]
            assert name in order_form
            query = order_form[name].order(query, value)
        return query

    def _page_query(self, env, query, page=1, page_size=1):
        assert page > 0
        assert 0 < page_size <= self.max_limit
        query = query.limit(page_size)
        if page != 1:
            query = query.offset((page-1)*page_size)
        return query


class I18nMixin:

    langs = ['ru', 'en']

    def __init__(self, component, lang, **kwargs):
        assert lang in self.langs
        self.lang = lang
        super().__init__(component, **kwargs)

    def get_id(self, **kwargs):
        lang = kwargs.get('lang', self.lang)
        id = super().get_id(**kwargs)
        return '.'.join([lang, id])

    @classmethod
    def create(cls, component, registry, **kwargs):
        assert 'lang' not in kwargs
        for lang in cls.langs:
            super().create(component, registry, lang=lang, **kwargs)

    @cached_property
    def i18n_streams(self):
        return {lang: self.streams[self.get_id(lang=lang)] \
                for lang in self.langs}

    @cached_property
    def mapper(self):
        mappers = self.component.app.db.mappers
        return mappers[self.db_id][self.lang][self.mapper_name]


class PublicationMixin:

    db_ids = ['admin', 'front']
    db_id = None
    allowed_perms = ['rwxdcp', 'rx']

    def __init__(self, component, db_id, **kwargs):
        assert db_id in self.db_ids
        self.db_id = db_id
        self.is_src = db_id == self.db_ids[0]
        self.is_dest = db_id == self.db_ids[1]
        super().__init__(component, **kwargs)
        self.set_permissions()

    def get_id(self, **kwargs):
        db_id = kwargs.get('db_id', self.db_id)
        id = super().get_id(**kwargs)
        return '.'.join([db_id, id])

    @classmethod
    def create(cls, component, registry, **kwargs):
        assert 'db_id' not in kwargs
        for db_id in cls.db_ids:
            super().create(component, registry, db_id=db_id, **kwargs)

    @cached_property
    def pub_streams(self):
        return {db_id: self.streams[self.get_id(db_id=db_id)] \
                for db_id in self.db_ids}

    def set_permissions(self):
        allowed_perms = set(self.allowed_perms[self.db_ids.index(self.db_id)])
        self.permissions = {role: set(perms).intersection(allowed_perms) \
            for role, perms in self.permissions.items()}


class I18nStream(I18nMixin, Stream):
    pass

class PubStream(PublicationMixin, Stream):
    pass

class PubI18nStream(PublicationMixin, I18nMixin, Stream):
    pass



from .. import base


class Component(base.Component):

    name = 'cache'

    async def get(self, key):
        raise NotImplementedError

    async def set(self, key, value, expire=0):
        raise NotImplementedError

    async def delete(self, key):
        raise NotImplementedError





from .generator import FullName
from .generator import FirstName
from .generator import LastName
from .generator import Image
from .generator import Date
from .generator import Text
from .generator import NumberedTitle
from .generator import EHTMLText
from .generator import Words
from .generator import Slug
from .generator import Constant
from .generator import Choice
from .generator import TrueOrFalse
from .generator import Relation

from .generator import Generator




from ikcms.ws_apps.base import exceptions

ClientError = exceptions.ClientError

class AccessDeniedError(exceptions.BaseError):
    message = 'Access Denied'


class InvalidCredentialsError(exceptions.BaseError):
    message = 'Invalid credentials'


class InvalidPasswordError(InvalidCredentialsError):
    message = 'Invalid login or password'


class InvalidTokenError(InvalidCredentialsError):
    message = 'Invalid token'



# -*- coding: utf-8 -*-
import sqlalchemy.orm
from iktomi.unstable.db.sqla.files import filesessionmaker

import ikcms.components.db.sqla


def sessionmaker(component, *args, **kwargs):
    app = component.app
    return filesessionmaker(
        sqlalchemy.orm.sessionmaker(*args, **kwargs),
        file_manager=app.admin_file_manager,
        file_managers={
            component.models['admin'].metadata: app.admin_file_manager,
            component.models['front'].metadata: app.front_file_manager,
        }
    )

Component = ikcms.components.db.sqla.component(
    session_maker_class=sessionmaker,
)


component = Component.create_cls




import asyncio

import ikcms.ws_apps.base

from . import exceptions


class App(ikcms.ws_apps.base.App):

    components = []

    def __init__(self, cfg, loop=None):
        super().__init__(cfg)
        if loop is None:
            loop = asyncio.get_event_loop()
        self.init_components(loop)

    def init_components(self, loop):
        if not self.components:
            return
        creators = [component.create(self) for component in self.components]
        names = [component.name for component in self.components]
        results, _ = loop.run_until_complete(asyncio.wait(creators))
        results = [result.result() for result in results]
        # save components order
        results_dict = {result.name: result for result in results}
        self.components = [results_dict[name] for name in names]

    def get_client_class(self):
        from .client import Client
        return Client

    def get_component(self, name):
        for component in self.components:
            if component.name == name:
                return component
        raise exceptions.ComponentNotFoundError(name)




from iktomi.cli.base import Cli as _Cli

class Cli(_Cli):
    name = None

    def __init__(self, App, Cfg):
        assert self.name
        self.App = App
        self.Cfg = Cfg

    def create_cfg(self, custom_cfg_path=None, **kwargs):
        cfg = self.Cfg()
        cfg.update_from_py(custom_cfg_path)
        cfg.update(kwargs)
        return cfg

    def create_app(self, **kwargs):
        cfg = self.create_cfg(**kwargs)
        return self.App(cfg)





from ikcms.forms.exceptions import RawValueError


__all__ = (
    'RawValueError',
    'ClientError'
    'BaseError',
    'ClientNotFoundError'
    'MessageError',
    'MessageError',
    'HandlerNotAllowedError',
)


class ClientError(Exception):

    def __init__(self, exc):
        self.exc = exc
        self.error = exc.error
        self.kwargs = exc.kwargs.copy()
        self.message = str(exc)

    def __str__(self):
        return self.message


class BaseError(Exception):

    message = 'Error message'
    kwargs = {}

    def __init__(self, **kwargs):
        self.kwargs = kwargs.copy()

    @property
    def error(self):
        return self.__class__.__name__

    def __str__(self):
        return self.message.format(**self.kwargs)


class ClientNotFoundError(BaseError):

    message = 'Client not found: client_id="{client_id}"'

    def __init__(self, client_id):
        super().__init__(client_id=client_id)

class ClientAlreadyAddedError(BaseError):

    message = 'Client already added: client_id="{client_id}"'

    def __init__(self, client_id):
        super().__init__(client_id=client_id)


class ProtocolError(BaseError): pass


class JSONDecodeError(ProtocolError):

    message = 'Expecting value: line {line} column {column} (char {pos})'

    def __init__(self, line, column, pos):
        super().__init__(line=line, column=column, pos=pos)


class RequestTypeError(ProtocolError):

    message = 'The request type must be "{required_type}", not "{current_type}"'

    def __init__(self, required_type, current_type):
        super().__init__(required_type=required_type, current_type=current_type)


class MessageError(ProtocolError):

    message = "{}"

    def __init__(self, errors):
        assert isinstance(errors, dict)
        super().__init__(errors=errors)

    def __str__(self):
        errors = ['Field "{}" error: {}.'.format(key, error) \
            for key, error in self.kwargs['errors'].items()]
        return self.message.format(' '.join(errors))


class HandlerNotAllowedError(BaseError):

    message = 'Handler Not Allowed: "{handler}"'

    def __init__(self, handler):
        super().__init__(handler=handler)





from unittest import TestCase
from unittest import skipIf
from unittest.mock import MagicMock
from datetime import date

from sqlalchemy import sql
#from sqlalchemy.testing.assertions import AssertsCompiledSQL

from ikcms.ws_components.streams import actions
from ikcms.ws_components.streams import exceptions

from ikcms import orm
from ikcms.utils.asynctests import asynctest
import ikcms.ws_components.db

from ikcms.ws_components.streams.streams import Stream
from ikcms.ws_components.streams.forms import list_fields
from ikcms.ws_components.streams.forms import filter_fields
from ikcms.ws_components.streams.forms import item_fields

from tests.models import create_models1, create_models2, create_metadata
from tests.cfg import cfg


def _items_from_keys(items, keys):
    return [dict([(key, value) \
            for key, value in item.items() if key in keys]) \
                for item in items]


@skipIf(not cfg.AIO_DB_ENABLED, 'db url undefined')
class ActionsTestCase(TestCase):
    item1 = {
        'id': 1,
        'title': '111t',
        'title2': '111t2',
        'date': date(2005, 4, 20),
    }
    item2 = {
        'id': 2,
        'title': '222t',
        'title2': '222t2',
        'date': date(1998, 7, 6),
    }
    item3 = {
        'id': 3,
        'title': '333t',
        'title2': '333t2',
        'date': date(2016, 12, 12),
    }
    item4 = {
        'id': 4,
        'title': '444t',
        'title2': '444t2',
        'date': date(2020, 1, 17),
    }
    items = [item1, item2, item3, item4]
    edit_items = _items_from_keys(items, ['id', 'title', 'date'])

    raw_item1 = {
        'id': 1,
        'title': '111t',
        'title2': '111t2',
        'date': '2005-04-20',
    }
    raw_item2 = {
        'id': 2,
        'title': '222t',
        'title2': '222t2',
        'date': '1998-07-06',
    }
    raw_item3 = {
        'id': 3,
        'title': '333t',
        'title2': '333t2',
        'date': '2016-12-12',
    }
    raw_item4 = {
        'id': 4,
        'title': '444t',
        'title2': '444t2',
        'date': '2020-01-17',
    }
    raw_items = [raw_item1, raw_item2, raw_item3, raw_item4]
    raw_edit_items = _items_from_keys(raw_items, ['id', 'title', 'date'])



    items_table_keys = {'id', 'title', 'title2', 'date'}
    items_relation_keys = set()
    items_allowed_keys = items_table_keys.union(items_relation_keys)

    models1 = create_models1()
    models2 = create_models2()
    mapper_cls = orm.mappers.Base

    async def asetup(self):
        registry = orm.mappers.Registry(
            create_metadata([self.models1, self.models2]))
        self.mapper_cls.from_model(registry, [self.models1.Test])

        app = MagicMock()
        app.cfg.DATABASES = {
            'db1': cfg.DB_URL,
            'db2': cfg.DB_URL,
        }
        del app.db
        db = await ikcms.ws_components.db.component(mappers=registry).create(app)
        async with await db() as session:
            conn1 = await session.get_connection(db.engines['db1'])
            await self.models1.reset(conn1)
            conn2 = await session.get_connection(db.engines['db2'])
            await self.models2.reset(conn2)

        class l_id(list_fields.id):
            widget = MagicMock()

        class l_title(list_fields.title):
            widget = MagicMock()

        class l_date(list_fields.Date):
            name = 'date'
            title = 'date'
            widget = MagicMock()

        class f_id(filter_fields.id):
            widget = MagicMock()

        class f_title(filter_fields.title):
            widget = MagicMock()

        class f_date(filter_fields.Date):
            name = 'date'
            title = 'date'
            widget = MagicMock()

        class f_title2(filter_fields.title):
            name = 'title2'
            title = 'title2'
            widget = MagicMock()

        class i_id(item_fields.id):
            widget = MagicMock()

            def get_initials(_self, **kwargs):
                return 50000

        class i_title(item_fields.title):
            widget = MagicMock()

            def get_initials(_self, **kwargs):
                test_kwarg = kwargs.get('test_kwarg', 'test_default')
                return '{}-{}-initials'.format(_self.name, test_kwarg)

        class i_date(item_fields.Date):
            name = 'date'
            title = 'date'
            widget = MagicMock()

            def get_initials(_self, **kwargs):
                return date(2005, 5, 5)


        class TestStream(Stream):
            max_limit = 50
            name = 'test_stream'
            title = 'test_stream_title'
            mapper_name = 'Test'
            db_id = 'db1'
            permissions = {'test_role': 'rwxcd'}

            list_fields = [
                l_id,
                l_title,
                l_date,
            ]
            filter_fields = [
                f_id,
                f_title,
                f_title2,
                f_date,
            ]
            item_fields = [
                i_id,
                i_title,
                i_date,
            ]

            def get_item_form(self, env, item=None, kwargs=None):
                kwargs = kwargs or {}
                raise_kwarg = kwargs.get('raise')
                if raise_kwarg:
                    raise raise_kwarg
                return super().get_item_form(env, item, kwargs)

            def check_perms(self, user, perms):
                pass


        stream = TestStream(MagicMock(app=app))
        env = MagicMock()
        env.app = app

        return {
            'db': db,
            'stream': stream,
            'env': env,
        }

    async def aclose(self, db, stream, env):
        await db.close()


    def _base_assert_list_action_response(self, resp, action, stream):
        self.assertEqual(resp['stream'], 'test_stream')
        self.assertEqual(resp['title'], 'test_stream_title')
        self.assertEqual(resp['action'], action.name)
        self.assertEqual(
            resp['list_fields'],
            [f.widget.to_dict(f) for f in stream.list_fields],
        )
        self.assertEqual(
            resp['filters_fields'],
            [f.widget.to_dict(f) for f in stream.filter_fields],
        )

    async def _set_table_state(self, db, table, state):
        async with await db() as session:
            query = sql.insert(table).values(state)
            result = await session.execute(query)

    async def _assert_table_state(self, db, table, state, keys=None):
        if keys:
            columns = [table.c[key] for key in keys]
        else:
            columns = table.c
        async with await db() as session:
            query = sql.select(columns).order_by(table.c.id)
            result = await session.execute(query)
            db_state = [dict(row) for row in result]
            self.assertEqual(state, db_state)

    @asynctest
    async def test_list(self, db, stream, env):
        action = actions.List(stream)
        mapper = db.mappers['db1']['Test']
        await self._set_table_state(db, self.models1.test_table1, self.items)

        # test order
        order_values = [
            (None, self.raw_edit_items),
            ('+id', self.raw_edit_items),
            ('+title', self.raw_edit_items),
            ('-title', self.raw_edit_items[::-1]),
        ]
        filters_values = [
            (None, lambda i: True),
            ({}, lambda i: True),
        ]
        for id in [-5, 0, 1, 3, 10]:
            def func(x):
                return lambda i: i['id'] == x
            filters_values.append(({'id': id}, func(id)))
        for title in ['1', 't', '2-333t', '24']:
            def func(x):
                return lambda i: i['title'].find(x) != -1
            filters_values.append(({'title': title}, func(title)))
        for order, raw_items in order_values:
            _order = order or '+id'
            for filters, filter_func in filters_values:
                _filters = filters or {}
                _raw_items = [i for i in raw_items if filter_func(i)]
                for page_size in [None, 1, 3, 4, 50]:
                    _page_size = page_size or 1
                    for page in [None, 1, 2, 3, 4, 10]:
                        _page = page or 1

                        kwargs = {}
                        if order is not None:
                            kwargs['order'] = order
                        if filters is not None:
                            kwargs['filters'] = filters
                        if page_size is not None:
                            kwargs['page_size'] = page_size
                        if page is not None:
                            kwargs['page'] = page

                        resp = await action.handle(env, kwargs)
                        self._base_assert_list_action_response(
                            resp, action, stream)
                        self.assertEqual(
                            resp['items'],
                            _raw_items[(_page-1)*_page_size:_page*_page_size],
                            {
                                'order': order,
                                'filters': filters,
                                'page_size': page_size,
                                'page': page,
                            }
                        )
                        self.assertEqual(resp['filters_errors'], {})
                        self.assertEqual(resp['filters'], _filters)
                        self.assertEqual(resp['page_size'], _page_size)
                        self.assertEqual(resp['page'], _page)
                        self.assertEqual(resp['order'], _order)
                        self.assertEqual(resp['total'], len(_raw_items))


        error_page_values = [-10, 0, 5.6, 'aaaa', '20', None]
        for value in error_page_values:
            with self.assertRaises(exceptions.ClientError) as ctx:
                await action.handle(env, {
                    'page': value,
                })
            exc = ctx.exception
            self.assertEqual(list(exc.kwargs['errors'].keys()), ['page'])

        error_page_size_values = [-10, 0, 5.6, 'aaa', '20', None]
        for value in error_page_size_values:
            with self.assertRaises(exceptions.ClientError) as ctx:
                await action.handle(env, {
                    'page': value,
                })
            exc = ctx.exception
            self.assertEqual(list(exc.kwargs['errors'].keys()), ['page'])

        with self.assertRaises(exceptions.ClientError) as ctx:
            await action.handle(env, {
                'filters': 56,
            })
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['filters'])
        with self.assertRaises(exceptions.ClientError) as ctx:
            await action.handle(env, {
                'order': {},
            })
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['order'])
        with self.assertRaises(exceptions.ClientError) as ctx:
            await action.handle(env, {
                'page': 'xxx',
            })
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['page'])
        with self.assertRaises(exceptions.ClientError) as ctx:
            await action.handle(env, {
                'page_size': 'xxx',
            })
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['page_size'])
        with self.assertRaises(exceptions.ClientError) as ctx:
            await action.handle(env, {
                'page_size': -5,
            })
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['page_size'])
        with self.assertRaises(exceptions.ClientError) as ctx:
            await action.handle(env, {
                'order': '+error_field',
            })
        exc = ctx.exception
        self.assertEqual(
            exc.kwargs,
            {'stream_name': 'test_stream', 'field_name': 'error_field'},
        )
        with self.assertRaises(exceptions.ClientError) as ctx:
            await action.handle(env, {
                'page_size': 100,
            })
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['page_size'])

        #XXX to do: test ValidationError


    @asynctest
    async def test_get_item(self, db, stream, env):
        action = actions.GetItem(stream)
        mapper = db.mappers['db1']['Test']
        await self._set_table_state(db, self.models1.test_table1, self.items)

        for raw_item in self.raw_edit_items:
            resp = await action.handle(env, {'item_id': raw_item['id']})
            self.assertEqual(
                resp['item_fields'],
                [f.widget.to_dict(f) for f in stream.item_fields],
            )
            self.assertEqual(resp['item'], raw_item)

        for value in [-10, 0, 5, 500]:
            with self.assertRaises(exceptions.ClientError):
                await action.handle(env, {'item_id': value})

        with self.assertRaises(exceptions.ClientError) as ctx:
            await action.handle(env, {})
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors']), ['item_id'])

        for value in [{}, None, [1, 2]]:
            with self.assertRaises(exceptions.ClientError) as ctx:
                await action.handle(env, {'item_id': value})
            ctx.exception


    @asynctest
    async def test_new_item(self, db, stream, env):
        action = actions.NewItem(stream)

        resp = await action.handle(env, {})
        self.assertEqual(
            resp['item_fields'],
            [f.widget.to_dict(f) for f in stream.item_fields],
        )
        self.assertEqual(
            resp['item'],
            {
                'id': 50000,
                'title': 'title-test_default-initials',
                'date': '2005-05-05',
            },
        )
        #test initials
        resp = await action.handle(env, {'kwargs': {'test_kwarg': 'test_init'}})
        self.assertEqual(
            resp['item_fields'],
            [f.widget.to_dict(f) for f in stream.item_fields],
        )
        self.assertEqual(
            resp['item'],
            {
                'id': 50000,
                'title': 'title-test_init-initials',
                'date': '2005-05-05',
            },
        )

        #test kwargs
        test_exc = Exception('test')
        with self.assertRaises(Exception) as ctx:
            resp = await action.handle(
                env,
                {'kwargs': {'raise': test_exc}}
            )
        self.assertEqual(ctx.exception, test_exc)


        for value in [None, 'xxx', [], 10]:
            with self.assertRaises(exceptions.ClientError) as ctx:
                await action.handle(env, {'kwargs': value})
            self.assertEqual(
                list(ctx.exception.kwargs['errors']),
                ['kwargs'],
            )

    @asynctest
    async def test_create_item(self, db, stream, env):
        mapper = db.mappers['db1']['Test']
        action = actions.CreateItem(stream)
        for item in self.raw_edit_items[:2]:
            _item = item.copy()
            _item.pop('id')
            resp = await action.handle(
                env,
                {'values': _item}
            )
            self.assertEqual(
                resp['item_fields'],
                [f.widget.to_dict(f) for f in stream.item_fields],
            )
            self.assertEqual(resp['item'], item)
            self.assertEqual(resp['errors'], {})

        for item in self.raw_edit_items[:-3:-1]:
            resp = await action.handle(env, {'values': item})
            self.assertEqual(
                resp['item_fields'],
                [f.widget.to_dict(f) for f in stream.item_fields],
            )
            self.assertEqual(resp['item'], item)
            self.assertEqual(resp['errors'], {})

        await self._assert_table_state(
            db,
            self.models1.test_table1,
            self.edit_items,
            keys=['id', 'title', 'date'],
        )

        # validation error
        _item = item.copy()
        _item['date'] = 'validation error'
        resp = await action.handle(
            env,
            {'values': _item}
        )
        self.assertEqual(
            resp['item_fields'],
            [f.widget.to_dict(f) for f in stream.item_fields],
        )
        self.assertEqual(resp['item'], _item)
        self.assertEqual(
            resp['errors'],
            {'date': stream.item_fields[2].conv.error_not_valid},\
        )

        # errors
        with self.assertRaises(exceptions.ClientError):
            await action.handle(env, {})

        with self.assertRaises(exceptions.ClientError):
            await action.handle(env, {'values': []})

        with self.assertRaises(exceptions.ClientError):
            await action.handle(env, {'kwargs': None})

        with self.assertRaises(exceptions.ClientError):
            resp = await action.handle(env, {'values': {'id': ''}})

        #test kwargs
        test_exc = Exception('test')
        with self.assertRaises(Exception) as ctx:
            resp = await action.handle(
                env,
                {'values': _item, 'kwargs': {'raise': test_exc}}
            )
        self.assertEqual(ctx.exception, test_exc)

    @asynctest
    async def test_update_item(self, db, stream, env):
        action = actions.UpdateItem(stream)
        mapper = db.mappers['db1']['Test']
        await self._set_table_state(db, self.models1.test_table1, self.items)
        items = self.edit_items.copy()

        resp = await action.handle(
            env,
            {
                'item_id': 3,
                'values': {'title': 'updated_title'}
            }
        )
        self.assertEqual(resp['item_id'], 3)
        self.assertEqual(
            resp['item_fields'],
            [f.widget.to_dict(f) for f in stream.item_fields],
        )
        self.assertEqual(
            resp['values'],
            {
                'title': 'updated_title',
            },
        )
        self.assertEqual(resp['errors'], {})
        items[3-1] = dict(items[3-1], title='updated_title')


        resp = await action.handle(
            env,
            {
                'item_id': 4,
                'values': {'id': 50, 'title': 'updated_title2'}
            }
        )
        self.assertEqual(resp['item_id'], 50)
        self.assertEqual(
            resp['item_fields'],
            [f.widget.to_dict(f) for f in stream.item_fields],
        )
        self.assertEqual(
            resp['values'],
            {
                'id': 50,
                'title': 'updated_title2',
            },
        )
        self.assertEqual(resp['errors'], {})
        items[4-1] = dict(
            items[4-1], id=50, title='updated_title2')

        await self._assert_table_state(
            db,
            self.models1.test_table1,
            items,
            keys=['id', 'title', 'date'],
        )

        # test errors
        with self.assertRaises(exceptions.ClientError) as ctx:
            resp = await action.handle(env, {})
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['item_id'])

        with self.assertRaises(exceptions.ClientError) as ctx:
            resp = await action.handle(env, {'item_id': 16})
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['values'])

        with self.assertRaises(exceptions.ClientError) as ctx:
            resp = await action.handle(env, {'item_id': 'error type'})
        exc = ctx.exception
        self.assertEqual(list(exc.kwargs['errors'].keys()), ['item_id'])

        with self.assertRaises(exceptions.ClientError):
            resp = await action.handle(
                env,
                {
                    'item_id': 11,
                    'values': {'id': 51, 'title': 'new_title'},
                }
            )

        # validation error
        resp = await action.handle(
            env,
            {
                'item_id': 11,
                'values': {'id': 51, 'date': 'validation error'},
            }
        )
        self.assertEqual(
            resp['item_fields'],
            [f.widget.to_dict(f) for f in stream.item_fields],
        )
        self.assertEqual(
            resp['errors'],
            {'date': stream.item_fields[2].conv.error_not_valid},\
        )


    @asynctest
    async def test_delete_item(self, db, stream, env):
        action = actions.DeleteItem(stream)
        mapper = db.mappers['db1']['Test']
        await self._set_table_state(db, self.models1.test_table1, self.items)

        resp = await action.handle(env, {'item_id': 3})
        self.assertEqual(resp['item_id'], 3)
        await self._assert_table_state(
            db,
            self.models1.test_table1,
            [self.item1, self.item2, self.item4],
        )

        with self.assertRaises(exceptions.ClientError):
            await action.handle(env, {})

        invalid_values = [None, 'aaa', [], {}, set()]
        for value in invalid_values:
            with self.assertRaises(exceptions.ClientError):
                await action.handle(env, {'item_id': value})

        with self.assertRaises(exceptions.ClientError):
            await action.handle(env, {'item_id': 500})





from . import exc

class Session:

    def __init__(self, engines, binds, **kwargs):
        self.engines = engines
        self.binds = binds
        self.connections = {}
        self.transactions = {}
        self.__dict__.update(kwargs)

    def get_engine(self, query):
        if hasattr(query, 'table'):
            # Insert, Update, Delete
            table = query.table
        elif hasattr(query, '_froms'):
            # Select
            table = list(query._froms)[0]
        else:
            raise exc.OrmError("Can't get query table")
        return self.binds[table]

    async def execute(self, query):
        conn = await self.get_connection(self.get_engine(query))
        return await conn.execute(query)

    async def get_connection(self, engine):
        conn = self.connections.get(engine)
        if conn:
            return conn
        self.connections[engine] = await engine.acquire()
        await self._begin(engine)
        return self.connections[engine]

    async def close(self):
        for engine in list(self.transactions):
            await self._rollback(engine)
        for engine, conn in self.connections.items():
            engine.release(conn)

    async def commit(self):
        for engine in list(self.transactions):
            await self._commit(engine)
            await self._begin(engine)

    async def rollback(self):
        for engine in list(self.transactions):
            await self._rollback(engine)
            await self._begin(engine)

    async def _begin(self, engine):
        assert engine not in self.transactions
        self.transactions[engine] = await self.connections[engine].begin()

    async def _commit(self, engine):
        assert engine in self.transactions
        await self.transactions[engine].commit()
        del self.transactions[engine]

    async def _rollback(self, engine):
        assert engine in self.transactions
        await self.transactions[engine].rollback()
        del self.transactions[engine]

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, e, tb):
        if exc_type:
            await self.close()
        else:
            for engine in list(self.transactions):
                await self._commit(engine)
            await self.close()




__all__ = (
    'Component',
)


class Component:

    name = None

    def __init__(self, app):
        assert self.name
        assert not hasattr(app, self.name)
        self.app = app
        self.handlers = self.collect_handlers()
        setattr(app, self.name, self)
        app.handlers.update(self.handlers)

    @classmethod
    async def create(cls, app):
        return cls(app)

    def collect_handlers(self):
        prefix = lambda name: '{}.{}'.format(self.name, name.split('_')[1])
        return {
            prefix(name): getattr(self, name)
            for name in dir(self) if name.startswith('h_')
        }

    def client_init(self, env):
        pass

    def client_close(self, env):
        pass

    @classmethod
    def create_cls(cls, **kwargs):
        return type(cls.__name__, (cls,), kwargs)




import os
import fnmatch
import sys
from multiprocessing import Process
from itertools import chain
from iktomi.cli.app import App as _AppCli
from iktomi.cli.app import wait_for_code_change
from iktomi.cli.app import flush_fds
from iktomi.cli.app import MAXFD
from .base import Cli
import logging
import time
from itertools import chain
import logging

logging.basicConfig()
logger = logging.getLogger()


def iter_module_files():
    for module in sys.modules.values():
        filename = getattr(module, '__file__', None)
        if filename:
            while not os.path.isfile(filename): # pragma: no cover
                # NOTE: this code is needed for the cases of importing
                # from archive or custom importers
                # for example, if we importing from archive foo.zip
                # module named zipped, then this zipped.__file__ will equal
                # to foo.zip/zipped.py, and os.path.dirname will give us
                # file, not directory.
                # It is marked as pragma: no cover, because this code was taken
                # from werkzeug and we believe that it is tested
                filename = os.path.dirname(filename)
                if not filename:
                    break
            else:
                if filename.endswith(('.pyc', '.pyo')):
                    filename = filename[:-1]
                yield filename

def wait_for_code_change(extra_files=None, interval=1):
    mtimes = {}
    while 1:
        for filename in chain(iter_module_files(), extra_files or ()):
            try:
                mtime = os.stat(filename).st_mtime
            except OSError: # pragma: no cover
                # this is cannot be guaranteed covered by coverage because of interpreter optimization
                # see https://bitbucket.org/ned/coveragepy/issues/198/continue-marked-as-not-covered#comment-4052311

                continue

            old_time = mtimes.get(filename)
            if old_time is None:
                mtimes[filename] = mtime
            elif mtime > old_time:
                logger.info('Changes in file "%s"', filename)
                return
        time.sleep(interval)

def http_process(host, port, stdin_fdno, app_factory):
    sys.stdin = os.fdopen(stdin_fdno)
    from wsgiref.simple_server import make_server
    app = app_factory.create_app()
    host = host or app.cfg.HTTP_SERVER_HOST
    port = port and int(port) or app.cfg.HTTP_SERVER_PORT
    print('Staring HTTP server {}:{}...'.format(host, port))
    server = make_server(host, port, app)
    server.serve_forever()


class AppCli(Cli):
    name = 'app'

    def command_serve(self, host=None, port=None, cfg=''):

        stdin = os.dup(sys.stdin.fileno())
        p1 = Process(target=http_process, args=(host, port, stdin, self))
        p1.start()

        cfg = self.create_cfg(custom_cfg_path=cfg)

        extra_files = []
        file_types = ['*.py', '*.yaml']

        for root, dirnames, filenames in os.walk(cfg.ROOT_DIR):
            filenames_to_check = chain.from_iterable(
                fnmatch.filter(filenames, files) for files in file_types
            )
            for filename in filenames_to_check:
                extra_files.append(os.path.join(root, filename))

        try:
            wait_for_code_change(extra_files=extra_files)
            p1.terminate()
            p1.join()
            flush_fds()

            pid = os.fork()
            if pid:
                os.closerange(3, MAXFD)
                os.waitpid(pid, 0)
                os.execvp(sys.executable, [sys.executable] + sys.argv)
            else:
                sys.exit()

        except KeyboardInterrupt:
            print('Terminating HTTP server...')
            p1.terminate()

        sys.exit()

    def command_shell(self, level=None, cfg=''):
        kwargs = dict(custom_cfg_path=cfg)
        if level:
            kwargs['LOG_LEVEL'] = level
        app = self.create_app(**kwargs)
        return self._cli(app).command_shell()

    def shell_namespace(self, app):
        return {
            'app': app,
        }

    def _cli(self, app):
        return _AppCli(app, shell_namespace=self.shell_namespace(app))



# -*- coding: utf-8 -*-
from iktomi.utils import cached_property


class Context(object):
    def __init__(self, env):
        self.env = env

    @cached_property
    def top_menu(self):
        return self.env.app.get_top_menu(self.env)

    @cached_property
    def users(self):
        models = self.env.app.db.get_models('admin')
        return self.env.db.query(models.AdminUser).filter_by(active=True).all()



import iktomi.web
from iktomi.web import request_filter

from .. import exceptions

from .base import h_match
from .base import h_prefix
from .base import h_namespace
from .base import h_cases

from .domains import h_domain
from .domains import h_subdomain

__all__ = (
    'WebHandler',
    'request_filter',
    'h_not_found',
    'h_server_error',
    'h_match',
    'h_prefix',
    'h_namespace',
    'h_cases',
    'h_static_files',
    'h_domain',
    'h_subdomain',
    'h_app',
)

WebHandler = iktomi.web.WebHandler

@request_filter
def h_not_found(env, data, next_handler=None):
    return env.app.HTTPNotFound()

@request_filter
def h_server_error(env, data, next_handler=None):
    return env.app.HTTPInternalServerError()


class HStaticFiles(iktomi.web.static_files):

    def __init__(self, location, url, enabled=True):
        super(HStaticFiles, self).__init__(location, url)
        self.enabled = enabled

    def static_files(self, env, data):
        if self.enabled:
            return super(HStaticFiles, self).static_files(env, data)
        else:
            return None
    __call__ = static_files

h_static_files = HStaticFiles


class HApp(iktomi.web.WebHandler):

    def __init__(self, app):
        self.app = app

    def app(self, env, data):
        app_request = self.app.get_request(env.request.environ)
        app_env = self.app.get_env(app_request)
        app_env._route_state = env._route_state
        app_data = self.app.get_data()
        return self.app.handle(app_env, app_data)
    __call__ = app

    def _locations(self):
        return self.app.handler._locations()

h_app = HApp



import json
import datetime

import ikcms.ws_components.base
from iktomi.utils import cached_property


class Component(ikcms.ws_components.base.Component):

    locks = {}

    def env_close(self, env):
        for lock in self.get_locks_by_session_id(env.session_id):
            del self.locks[lock]

    def acquire(self, env, lock):
        session_id = self.locks.get(lock)
        if session_id:
            return bool(session_id == env.session_id)
        else:
            locks[lock] = env.session_id
            return True

    def take(self, env, lock):
        self.locks[lock] = env.session_id

    def release(self, env, lock):
        session_id = self.locks.get(lock)
        if session_id:
            if session_id == env.session_id:
                del self.locks[lock]
                return True
            else:
                return False
        else:
            return True

    def get_locks_by_session_id(self, session_id):
        return [id for id, sid in self.locks.items() \
                if sid==session_id]

    def _lock_name(self, *names):
        return '.'.join(names)


component = Component.create_cls



from collections import OrderedDict

from . import convs
from . import validators
from . import widgets
from . import exceptions


__all__ = (
    'NOTSET',
    'Base',
    'Field',
    'String',
    'Int',
    'Dict',
    'List',
    'RawDict',
    'RawList',
    'Block',
    'Date',
)


class NOTSET:
    pass


class Base:

    name = None
    label = None
    conv = convs.Converter
    fields = []
    validators = ()
    widget = widgets.Widget()

    raw_required = True
    not_none = True
    to_python_default = NOTSET

    def __init__(self, context=None, parent=None):
        self.context = context and context.copy() or {}
        self.parent = parent
        self.fields = [f(context, self) for f in self.fields]
        self.named_fields = OrderedDict(
            [(f.name, f) for f in self.fields if f.name])
        self.conv = self.conv(self)
        self.validators = [v(self) for v in self.validators]

    def to_python(self, raw_value):
        if raw_value is NOTSET:
            python_value = self._raw_value_notset()
        else:
            python_value = self.conv.to_python(raw_value)
        for v in self.validators:
            python_value = v(python_value)
        return python_value

    def from_python(self, python_value):
        if python_value is NOTSET:
            raise exceptions.PythonValueRequiredError(self.name)
        return self.conv.from_python(python_value)

    def _raw_value_notset(self):
        if self.raw_required:
            raise exceptions.RawValueRequiredError(self.name)
        return self.to_python_default

    def get_initials(self, **kwargs):
        return None



class Field(Base):

    def to_python(self, raw_dict):
        raw_value = raw_dict.get(self.name, NOTSET)
        try:
            python_value = super().to_python(raw_value)
        except exceptions.ValidationError as exc:
            raise exceptions.ValidationError({self.name: exc.kwargs['error']})
        if python_value is NOTSET:
            return {}
        else:
            return {self.name: python_value}

    def from_python(self, python_dict):
        python_value = python_dict.get(self.name, NOTSET)
        raw_value = super().from_python(python_value)
        return {self.name: raw_value}


class String(Field):
    conv = convs.Str
    validators = [
        validators.Required,
        validators.Regex,
        validators.Len,
    ]
    required = False
    regex = None
    min_len = None
    max_len = None


class Int(Field):
    conv = convs.Int
    validators = [
        validators.Required,
        validators.Range,
    ]
    required = False
    min_value = None
    max_value = None


class IntStr(Int):
    conv = convs.IntStr


class Dict(Field):
    conv = convs.Dict
    validators = [
        validators.Required,
    ]
    required = False


class List(Field):
    conv = convs.List
    validators = [
        validators.Required,
        validators.Len,
    ]
    required = False
    min_len = None
    max_len = None


class RawDict(Field):
    conv = convs.RawDict


class RawList(Field):
    conv = convs.RawList


class Block(Base):
    conv = convs.Dict
    validators = [
        validators.Required,
    ]
    required = False

    def to_python(self, raw_dict):
        raw_value = raw_dict.get(self.name, NOTSET)
        try:
            python_value = super().to_python(raw_value)
        except exceptions.ValidationError as exc:
            raise exceptions.ValidationError({self.name: exc.error})
        if python_value is NOTSET:
            return {}
        else:
            return python_value

    def from_python(self, python_dict):
        raw_value = super().from_python(python_dict)
        return {self.name: raw_value}


class Date(Field):
    conv = convs.Date
    format = "%Y-%m-%d"



from ikcms.ws_apps.base.forms import MessageForm as MessageFormBase
from ikcms import orm

from .forms import message_fields
from . import exceptions


class Base:
    name = None
    require_perms = ''

    MessageForm = MessageFormBase

    def __init__(self, stream):
        assert self.name is not None
        self.stream = stream

    async def handle(self, client, raw_message):
        self.stream.check_perms(client, self.require_perms)
        try:
            message = self.MessageForm().to_python_or_exc(raw_message)
        except exceptions.MessageError as exc:
            raise exceptions.ClientError(exc)
        return await self(client, message)

    async def _handle(self, env, message):
        raise NotImplementedError

    def get_cfg(self, client):
        return {
            'name': self.name
        }



class List(Base):
    name = 'list'
    require_perms = 'x'

    class MessageForm(MessageFormBase):
        fields = [
            message_fields.filters,
            message_fields.order,
            message_fields.page,
            message_fields.page_size,
        ]

    async def list(self, env, message):
        list_form = self.stream.get_list_form(env)
        filter_form = self.stream.get_filter_form(env)
        order_form = self.stream.get_order_form(env)
        order = message['order']
        if order[1:] not in order_form:
            raise exceptions.ClientError(
                exceptions.StreamFieldNotFound(self.stream.name, order[1:]),
            )

        raw_filters = message['filters']
        filters, filters_errors = filter_form.to_python(raw_filters)
        page = message['page']
        page_size = message['page_size']

        if not 0 < page_size <= self.stream.max_limit:
            raise exceptions.ClientError(
                exceptions.MessageError({'page_size':'Page size error'}),
            )
        async with await env.app.db() as session:
            list_items = await self.stream.list_items(
                env,
                session,
                filters,
                [order],
                page,
                page_size,
                keys=set(list_form.keys()),
            )
            total = await self.stream.count_items(env, session, filters)

        raw_list_items = list_form.values_from_python(list_items)

        return {
            'stream': self.stream.name,
            'title': self.stream.title,
            'action': self.name,
            'list_fields': list_form.get_cfg(),
            'items': raw_list_items,
            'total': total,
            'filters_fields': filter_form.get_cfg(),
            'filters_errors': filters_errors,
            'filters': raw_filters,
            'page_size': page_size,
            'page': page,
            'order': message['order'],
        }
    __call__ = list


class GetItem(Base):

    name = 'get_item'
    require_perms = 'r'

    class MessageForm(MessageFormBase):
        fields = [
            message_fields.item_id,
        ]

    async def get_item(self, env, message):
        item_id = message['item_id']
        async with await env.app.db() as session:
            item = await self.stream.get_item(env, session, item_id)
        if not item:
            raise exceptions.ClientError(
                exceptions.StreamItemNotFoundError(self.stream.name, item_id)
            )
        item_fields_form = self.stream.get_item_form(env, item=item)
        raw_item = item_fields_form.from_python(item)
        return {
            'item_fields': item_fields_form.get_cfg(),
            'item': raw_item,
        }
    __call__ = get_item


class NewItem(Base):

    name = 'new_item'
    require_perms = 'c'

    class MessageForm(MessageFormBase):
        fields = [
            message_fields.kwargs,
        ]

    async def new_item(self, env, message):
        kwargs=message['kwargs']
        item_fields_form = self.stream.get_item_form(env, kwargs)
        item = await self.stream.new_item(env, kwargs)
        raw_item = item_fields_form.from_python(item)
        return {
            'item_fields': item_fields_form.get_cfg(),
            'item': raw_item,
        }
    __call__ = new_item


class CreateItem(Base):

    name = 'create_item'
    require_perms = 'c'

    class MessageForm(MessageFormBase):
        fields = [
            message_fields.values,
            message_fields.kwargs,
        ]

    async def create_item(self, env, message):
        item_fields_form = self.stream.get_item_form(
            env, kwargs=message['kwargs'])
        raw_item = message['values']
        keys = list(item_fields_form)
        if 'id' not in raw_item:
            keys.remove('id')
        try:
            item, errors = item_fields_form.to_python(raw_item, keys=keys)
        except exceptions.MessageError as exc:
            raise exceptions.ClientError(exc)
        if not errors:
            async with await env.app.db() as session:
                item = await self.stream.insert_item(env, session, item)
            raw_item = item_fields_form.from_python(item)
        return {
            'item_fields': item_fields_form.get_cfg(),
            'item': raw_item,
            'errors': errors,
        }
    __call__ = create_item


class UpdateItem(Base):

    name = 'update_item'
    require_perms = 'e'

    class MessageForm(MessageFormBase):
        fields = [
            message_fields.item_id,
            message_fields.values,
        ]

    async def update_item(self, env, message):
        item_fields_form = self.stream.get_item_form(env)
        item_id = message['item_id']
        raw_values = message['values']
        keys=raw_values.keys()
        values, errors = item_fields_form.to_python(raw_values, keys)
        if not errors:
            async with await env.app.db() as session:
                try:
                    item = await self.stream.update_item(
                        env,
                        session,
                        item_id,
                        values,
                    )
                except exceptions.StreamItemNotFoundError as exc:
                    raise exceptions.ClientError(exc)
        return {
            'item_fields': item_fields_form.get_cfg(),
            'item_id': raw_values.get('id', item_id),
            'values': raw_values,
            'errors': errors,
        }
    __call__ = update_item


class DeleteItem(Base):

    name = 'delete_item'
    require_perms = 'd'

    class MessageForm(MessageFormBase):
        fields = [
            message_fields.item_id,
        ]

    async def delete_item(self, env, message):
        async with await env.app.db() as session:
            try:
                await self.stream.delete_item(env, session, message['item_id'])
            except exceptions.StreamItemNotFoundError as exc:
                raise exceptions.ClientError(exc)
        return {
            'item_id': message['item_id'],
        }
    __call__ = delete_item
